from configparser import ConfigParser
from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.cycle_detection.half_cycle_detector import HalfCycleDetector
from simses.commons.cycle_detection.no_cycle_detector import NoCycleDetector
from simses.commons.cycle_detection.rainflow_cycle_detector import RainflowCycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.battery_management_system.management_system import BatteryManagementSystem
from simses.technology.lithium_ion.cell.generic import GenericCell
from simses.technology.lithium_ion.cell.isea import IseaCellType
from simses.technology.lithium_ion.cell.lfp_sony import SonyLFP
from simses.technology.lithium_ion.cell.lfp_sony_generic import SonyLFPGeneric
from simses.technology.lithium_ion.cell.lmo_daimler import DaimlerLMO
from simses.technology.lithium_ion.cell.lto_lmo import LTOLMO
from simses.technology.lithium_ion.cell.lto_nmc import LTONMC
from simses.technology.lithium_ion.cell.nca_panasonic_generic import PanasonicNCAGeneric
from simses.technology.lithium_ion.cell.nca_panasonic_ncr import PanasonicNCA
from simses.technology.lithium_ion.cell.nmc_molicel import MolicelNMC
from simses.technology.lithium_ion.cell.nmc_molicel_generic import MolicelNMCGeneric
from simses.technology.lithium_ion.cell.nmc_samsung78Ah import Samsung78AhNMC
from simses.technology.lithium_ion.cell.nmc_samsung94Ah import Samsung94AhNMC
from simses.technology.lithium_ion.cell.nmc_samsung94Ah_hybrid import Samsung94AhNMCHybrid
from simses.technology.lithium_ion.cell.nmc_sanyo_ur18650e import SanyoNMC
from simses.technology.lithium_ion.cell.nmc_lgmj1 import LGMJ1_NMC
from simses.technology.lithium_ion.cell.sodium_ion.sodium_ion_green_rock import SodiumIonGreenRock
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel
from simses.technology.lithium_ion.degradation.generic_cell import GenericCellDegradationModel
from simses.technology.lithium_ion.degradation.lfp_generic import LFPGenericDegradationModel
from simses.technology.lithium_ion.degradation.lfp_sony import SonyLFPDegradationModel
from simses.technology.lithium_ion.degradation.lmo_daimler import DaimlerLMODegradationModel
from simses.technology.lithium_ion.degradation.lto_lmo import LTOLMODegradationModel
from simses.technology.lithium_ion.degradation.lto_nmc import LTONMCDegradationModel
from simses.technology.lithium_ion.degradation.nca_generic import NCAGenericDegradationModel
from simses.technology.lithium_ion.degradation.nca_panasonicNCR import \
    PanasonicNCADegradationModel
from simses.technology.lithium_ion.degradation.nmc_generic import NMCGenericDegradationModel
from simses.technology.lithium_ion.degradation.nmc_molicel import MolicelNMCDegradationModel
from simses.technology.lithium_ion.degradation.nmc_samsung94Ah import \
    Samsung94AhNMCDegradationModel
from simses.technology.lithium_ion.degradation.nmc_sanyo_ur18650e import \
    SanyoNMCDegradationModel
from simses.technology.lithium_ion.degradation.nmc_lgmj1 import LGMJ1_NMCDegradationModel
from simses.technology.lithium_ion.degradation.no_degradation import NoDegradationModel
from simses.technology.lithium_ion.equivalent_circuit_model.equivalent_circuit import EquivalentCircuitModel
from simses.technology.lithium_ion.equivalent_circuit_model.rint_model import RintModel

class LithiumIonFactory:
    
    def __init__(self, config: ConfigParser):
        self.__log: Logger = Logger(type(self).__name__)
        self.__config_factory: StorageSystemConfig = StorageSystemConfig(config)
        self.__config_general: GeneralSimulationConfig = GeneralSimulationConfig(config)
        self.__config_battery: BatteryConfig = BatteryConfig(config)
        self.__config_battery_data: BatteryDataConfig = BatteryDataConfig()

    def create_battery_state_from(self, system_id: int, storage_id: int, cell_type: CellType,
                                  temperature: float, soc: float) -> LithiumIonState:
        time: float = self.__config_general.start
        bs = LithiumIonState(system_id, storage_id)
        bs.time = time
        bs.soc = soc
        bs.temperature = temperature
        bs.voltage = cell_type.get_open_circuit_voltage(bs)
        bs.voltage_open_circuit = cell_type.get_open_circuit_voltage(bs)
        bs.nominal_voltage = cell_type.get_nominal_voltage()

        bs.resistance_increase = cell_type.get_resistance_increase_start()
        bs.internal_resistance = cell_type.get_internal_resistance(bs) * (1 + bs.resistance_increase)
        bs.capacity = cell_type.get_capacity(bs) * cell_type.get_nominal_voltage() * cell_type.get_soh_start()
        bs.soh = cell_type.get_soh_start()
        bs.soe = bs.soc * bs.capacity
        bs.fulfillment = 1.0
        bs.max_charge_power = cell_type.get_max_current(bs) * bs.voltage
        bs.max_discharge_power = cell_type.get_min_current(bs) * bs.voltage
        return bs

    def create_cell_type(self, cell_type: str, voltage: float, capacity: float, soh: float) -> CellType:
        if cell_type.startswith(SonyLFPGeneric.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return SonyLFPGeneric(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(SonyLFP.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return SonyLFP(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(PanasonicNCAGeneric.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return PanasonicNCAGeneric(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(PanasonicNCA.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return PanasonicNCA(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(MolicelNMCGeneric.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return MolicelNMCGeneric(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(MolicelNMC.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return MolicelNMC(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(SanyoNMC.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return SanyoNMC(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(GenericCell.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return GenericCell(voltage, capacity, soh, self.__config_battery)
        elif cell_type.startswith(LTOLMO.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return LTOLMO(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(LTONMC.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return LTONMC(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(Samsung78AhNMC.__name__):
            self.__log.debug('Creating cell type as ' + cell_type)
            return Samsung78AhNMC(voltage, capacity, soh, self.__config_battery)
        elif cell_type.startswith(DaimlerLMO.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return DaimlerLMO(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(Samsung94AhNMCHybrid.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return Samsung94AhNMCHybrid(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(Samsung94AhNMC.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return Samsung94AhNMC(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(IseaCellType.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            try:
                file_name: str = cell_type.split(';')[1]
            except:
                raise Exception('Required file for '+ cell_type + ' could not be found')
            return IseaCellType(voltage, capacity, soh, self.__config_battery, self.__config_battery_data, file_name)
        elif cell_type.startswith(SodiumIonGreenRock.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return SodiumIonGreenRock(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        elif cell_type.startswith(LGMJ1_NMC.__name__):
            self.__log.debug('Creating cell type as' + cell_type)
            return LGMJ1_NMC(voltage, capacity, soh, self.__config_battery, self.__config_battery_data)
        else:
            options: [str] = list()
            options.append(SonyLFP.__name__)
            options.append(LTOLMO.__name__)
            options.append(LTONMC.__name__)
            options.append(PanasonicNCA.__name__)
            options.append(MolicelNMC.__name__)
            options.append(SanyoNMC.__name__)
            options.append(GenericCell.__name__)
            options.append(Samsung78AhNMC.__name__)
            options.append(DaimlerLMO.__name__)
            options.append(Samsung94AhNMC.__name__)
            options.append(Samsung94AhNMCHybrid.__name__)
            options.append(LGMJ1_NMC.__name__)
            raise Exception('Specified cell type ' + cell_type + ' is unknown. '
                            'Following options are available: ' + str(options))

    def create_cycle_detector(self, start_soc: float) -> CycleDetector:
        cycle_detector = self.__config_factory.cycle_detector
        if cycle_detector == HalfCycleDetector.__name__:
            self.__log.debug('Creating cycle detector as ' + cycle_detector)
            return HalfCycleDetector(start_soc, self.__config_general)
        elif cycle_detector == RainflowCycleDetector.__name__:
            self.__log.debug('Creating cycle detector as ' + cycle_detector)
            return RainflowCycleDetector(start_soc, self.__config_general, self.__config_battery)
        elif cycle_detector == NoCycleDetector.__name__:
            self.__log.debug('Creating cycle detector as ' + cycle_detector)
            return NoCycleDetector()
        else:
            self.__log.warn(
                'Specified cycle detector ' + str(cycle_detector) + ' not found, creating ' + NoCycleDetector.__name__)
            return NoCycleDetector()

    def create_degradation_model_from(self, cell_type: CellType, battery_state: LithiumIonState) -> DegradationModel:
        cycle_detector: CycleDetector = self.create_cycle_detector(battery_state.soc)
        if isinstance(cycle_detector, NoCycleDetector):
            self.__log.debug('Creating NoDegradationModel for cell type ' + cell_type.__class__.__name__)
            return NoDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, SonyLFPGeneric):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return LFPGenericDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, SonyLFP):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return SonyLFPDegradationModel(cell_type, cycle_detector, self.__config_battery, self.__config_battery_data)
        elif isinstance(cell_type, PanasonicNCAGeneric):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return NCAGenericDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, PanasonicNCA):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return PanasonicNCADegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, MolicelNMCGeneric):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return NMCGenericDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, MolicelNMC):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return MolicelNMCDegradationModel(cell_type, cycle_detector, self.__config_battery, self.__config_battery_data)
        elif isinstance(cell_type, SanyoNMC):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return SanyoNMCDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, DaimlerLMO):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return DaimlerLMODegradationModel(cell_type, cycle_detector, self.__config_battery, self.__config_battery_data)
        elif isinstance(cell_type, Samsung94AhNMCHybrid):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return SonyLFPDegradationModel(cell_type, cycle_detector, self.__config_battery, self.__config_battery_data)
        elif isinstance(cell_type, Samsung94AhNMC):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return Samsung94AhNMCDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, GenericCell):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return GenericCellDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, LTOLMO):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return LTOLMODegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, LTONMC):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return LTONMCDegradationModel(cell_type, cycle_detector, self.__config_battery)
        elif isinstance(cell_type, LGMJ1_NMC):
            self.__log.debug('Creating degradation model for cell type ' + cell_type.__class__.__name__)
            return LGMJ1_NMCDegradationModel(cell_type, cycle_detector, self.__config_battery)
        else:
            self.__log.warn('No degradation model found for cell type ' + cell_type.__class__.__name__)
            return NoDegradationModel(cell_type, cycle_detector, self.__config_battery)

    def create_battery_management_system_from(self, cell_type: CellType,
                                              battery_management_system: BatteryManagementSystem = None) -> BatteryManagementSystem:
        if battery_management_system is None:
            self.__log.debug('Creating lithium_ion management system for cell type ' + cell_type.__class__.__name__)
            return BatteryManagementSystem(cell_type, self.__config_battery)
        else:
            return battery_management_system

    def create_battery_model_from(self, cell_type: CellType,
                                  battery_model: EquivalentCircuitModel = None) -> EquivalentCircuitModel:
        if battery_model is None:
            self.__log.debug('Creating lithium_ion model for cell type ' + cell_type.__class__.__name__)
            return RintModel(cell_type)
        else:
            return battery_model

    def close(self):
        self.__log.close()



from abc import ABC, abstractmethod

from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType


class CalendarDegradationModel(ABC):
    """
    Degradation Model for the calendaric aging of the ESS.
    """

    def __init__(self, cell_type: CellType):
        super().__init__()
        self._cell: CellType = cell_type

    @abstractmethod
    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        """
        update the calendric capacity loss of a battery
        Attention: without considering internal aging variation the capacity loss calculation is conducted on a
        down-scaled single cell and again up-scaled to module or pack assuming that each single cell has the same
        degradation state
        Parameters
        ----------
            battery_state : LithiumIonState
            Current BatteryState of the lithium_ion. Used to determine if the lithium_ion is charging or discharging.

        Returns
        -------

        """
        pass

    @abstractmethod
    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        """
        update the calendric internal resistance increase of a battery
        Attention: without considering internal aging variation the resistance increase calculation is conducted on a
        down-scaled single cell and again up-scaled to module or pack assuming that each single cell has the same
        degradation state
        Parameters
        ----------
            battery_state : LithiumIonState
            Current BatteryState of the lithium_ion. Used to determine if the lithium_ion is charging or discharging.

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_degradation(self) -> float:
        """
        get the updated calendric capacity loss caused by the current calculation step (differential capacity loss)

        Returns
        -------
        float:
            differential capacity loss of the current step in [Ah]
        """
        pass

    @abstractmethod
    def get_resistance_increase(self) -> float:
        """
        get the updated calendric resistance increase til the current step (differential increase)
        Returns
        -------
        float:
            differential resistance increase til the current step in [p.u.]
        """
        pass

    @abstractmethod
    def reset(self, battery_state: LithiumIonState) -> None:
        """
        resets all values within a calendar degradation model, if battery is replaced
        Parameters
        ----------
            battery_state : LithiumIonState; Current BatteryState of the lithium_ion.

        Returns
        -------
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closing all resources in calendar degradation model

        Returns
        -------

        """
        pass



from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class NoCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self):
        super().__init__(None, None)

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        pass

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        pass

    def get_degradation(self) -> float:
        return 0

    def get_resistance_increase(self) -> float:
        return 0

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        pass

    def close(self) -> None:
        pass



from abc import ABC, abstractmethod

from simses.commons.state.technology.redox_flow import RedoxFlowState


class ElectrolyteSystem(ABC):
    """The ElectrolyteSystem describes the properties of the liquid storage medium of the redox flow battery."""

    def __init__(self, capacity: float):
        self.FARADAY = 96485  # As/mol
        self.__capacity = capacity

    def get_start_capacity(self) -> float:
        """
        Returns the total start capacity of the electrolyte of the redox flow system.

        Returns
        -------
        float:
            Start capacity in Wh.
        """
        return self.__capacity

    @abstractmethod
    def get_vanadium_concentration(self) -> float:
        """
        Returns the total vanadium concentration of the electrolyte.

        Returns
        -------
        float :
            total vanadium concentration of the electrolyte in mol/m^3.
        """
        pass

    @abstractmethod
    def get_viscosity_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the anolyte viscosity depending on state-of-charge (SOC) and temperature.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float:
            Analyte viscosity in Pas.
        """
        pass

    @abstractmethod
    def get_viscosity_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the anolyte viscosity depending on the state-of-charge (SOC) and temperature.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float:
            Catholyte viscosity in Pas.
        """
        pass

    @abstractmethod
    def get_min_viscosity(self) -> float:
        """
        Determines the minimal viscosity during operation of the redox flow battery.

        Returns
        -------
        float:
            Minimal viscosity in Pas
        """
        pass

    @abstractmethod
    def get_max_viscosity(self) -> float:
        """
        Determines the maximal viscosity during operation of the redox flow battery.

        Returns
        -------
        float:
            Maximal viscosity in Pas.
        """
        pass

    @abstractmethod
    def get_max_temperature(self) -> float:
        """
        Returns the maximal temperature of the electrolyte.

        Returns
        -------
        float:
            Maximal electrolyte temperature in K.
        """
        pass

    @abstractmethod
    def get_min_temperature(self) -> float:
        """
        Returns the minimal temperature of the electrolyte.

        Returns
        -------
        float:
            Minimal electrolyte temperature in K.
        """
        pass

    @abstractmethod
    def get_capacity_density(self) -> float:
        """
        Returns the volume specific capacity density of the electrolyte in As/m^3.

        Returns
        -------
        float:
            Volume specific capacity density in As/m^3.
        """
        pass

    def get_electrolyte_density(self) -> float:
        """
        Returns the mean density of the electrolyte in kg/m^3.

        Returns
        -------
        float:
            Density of the electrolyte in kg/m^3.
        """
        pass

    @abstractmethod
    def close(self):
        """Closing all resources in ElectrolyteSystem."""
        pass



from abc import ABC, abstractmethod

from simses.commons.state.abstract_state import State


class StorageTechnologyState(State, ABC):

    """
    A state with information specifically provided by a storage technology for each simulation timestep.
    """

    def __init__(self):
        super().__init__()

    @property
    @abstractmethod
    def is_charge(self) -> bool:
        """
        Returns True if current state presents a charging process, false otherwise

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def voltage(self) -> float:
        """
        DC Voltage in V

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def current(self) -> float:
        """
        DC Current in A

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def power_loss(self) -> float:
        """
        Power losses in W

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def soc(self) -> float:
        """
        SOC in p.u.

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def soh(self) -> float:
        """
        SOC in p.u.

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def capacity(self) -> float:
        """
        Capacity in Wh

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def fulfillment(self) -> float:
        """
        Stage of power fulfillment of storage data in p.u.

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def temperature(self) -> float:
        """
        Temperature of storage technology in K

        Returns
        -------

        """
        pass

    @temperature.setter
    def temperature(self, value: float) -> None:
        pass

    @property
    @abstractmethod
    def max_charge_power(self) -> float:
        """
        Current maximum charging power of storage technology in W

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def max_discharge_power(self) -> float:
        """
        Current maximum discharging power of storage technology in W

        Returns
        -------

        """
        pass



from configparser import ConfigParser

from simses.simulation.batch_processing import BatchProcessing


class FCRCaseStudyBatchProcessing(BatchProcessing):

    def __init__(self):
        super().__init__(do_simulation=True, do_analysis=True)

    def __generate_dc_systems(self) -> dict:
        config_set: dict = dict()
        config_set['lib_big'] = 'system_1,no_loss,lib_big\n'
        subsystems: str = ''
        for i in range(8):
            subsystems += 'system_1,fix,lib_small\n'
        config_set['lib_small'] = subsystems
        return config_set

    def __generate_ac_systems(self) -> dict:
        config_set: dict = dict()
        config_set['single'] = 'system_1,1.6e6,600,notton,no_housing,no_hvac\n'
        config_set['cascade'] = 'system_1,1.6e6,600,notton_cascade,no_housing,no_hvac\n'
        return config_set

    def _setup_config(self) -> dict:
        # generate config
        dc_systems: dict = self.__generate_dc_systems()
        ac_systems: dict = self.__generate_ac_systems()
        # setup config
        configs: dict = dict()
        for dc_name,dc_system in dc_systems.items():
            for ac_name,ac_system in ac_systems.items():
                config: ConfigParser = ConfigParser()
                config.add_section('STORAGE_SYSTEM')
                config.set('STORAGE_SYSTEM', 'STORAGE_SYSTEM_AC', ac_system)
                config.set('STORAGE_SYSTEM', 'STORAGE_SYSTEM_DC', dc_system)
                configs['simses_' + dc_name + '_' + ac_name] = config
        # print configs
        print(dc_systems)
        print('Configured ' + str(len(configs)) + ' simulations')
        return configs

    def clean_up(self) -> None:
        pass


if __name__ == "__main__":
    batch_processing: BatchProcessing = FCRCaseStudyBatchProcessing()
    batch_processing.run()
    print(type(batch_processing).__name__ + ' finished')



import math
import numpy as np
import scipy.interpolate

import pandas as pd
import scipy.interpolate

from simses.commons.config.data.redox_flow import RedoxFlowDataConfig
from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class IndustrialStack1500W(StackModule):
    """IndustrialStack1500W is a stack parameterised with field data of a commercially available redox flow system. The
    internal resistance considers an increase due to mass transport effects when the pump are off and the power exceeds
    a certain limit value."""

    __STACK_POWER_NOM = 1500  # W
    __CELL_NUMBER = 18  # -
    __CELL_AREA = 551  # cm^2
    __ELECTRODE_THICKNESS = 0.55  # cm
    __ELECTRODE_POROSITY = 0.93  # -
    __MIN_CELL_VOLTAGE = 1.22  # V
    __MAX_CELL_VOLTAGE = 1.55  # V
    __SELF_DISCHARGE_CURRENT = 0.96  # A
    __DEPENDENT_PARAMETERS = False

    def __init__(self, electrolyte_system: ElectrolyteSystem, voltage: float, power: float,
                 redox_flow_data_config: RedoxFlowDataConfig, redox_flow_config: RedoxFlowConfig):
        super().__init__(electrolyte_system, voltage, power, self.__CELL_NUMBER, self.__STACK_POWER_NOM,
                         redox_flow_config)
        self.__log: Logger = Logger(__name__)
        self.__SHUNT_FILE = redox_flow_data_config.industrial_stack_1500w_shunt_current
        self.__shunt_current = pd.read_csv(self.__SHUNT_FILE, header=None)  # A
        self.__shunt_current_mat = self.__shunt_current.iloc[1:, 1:]
        self.__soc_arr = self.__shunt_current.iloc[1:, 0]
        self.__current_arr = self.__shunt_current.iloc[0, 1:]
        self.__shunt_interp2d = scipy.interpolate.interp2d(self.__current_arr, self.__soc_arr, self.__shunt_current_mat,
                                                           kind='linear')
        self.__hydraulic_resistance = 5.39e10 * (self._SERIAL_SCALE * self._PARALLEL_SCALE) ** -0.9
        soc_arr = np.array(
            [0, 5.21, 9.63, 14.16, 19.88, 25.73, 30.55, 34.66, 39.67, 45.01, 49.93, 54.89, 59.9, 64.95, 69.98, 74.95,
             79.84, 84.63, 89.98, 95.4, 100.0])
        ocv_arr = np.array(
            [1.226, 1.258, 1.281, 1.30, 1.319, 1.336, 1.349, 1.361, 1.372, 1.383, 1.394, 1.405, 1.416, 1.428, 1.439,
             1.453, 1.463, 1.477, 1.493, 1.514, 1.537])
        self.__ocv_interp1d = scipy.interpolate.interp1d(soc_arr / 100, ocv_arr, kind='linear')

    def get_open_circuit_voltage(self, redox_flow_state: RedoxFlowState) -> float:
        soc_stack = redox_flow_state.soc_stack
        ocv_stack = self.__ocv_interp1d(soc_stack) * self.__CELL_NUMBER
        return ocv_stack * self.get_serial_scale()

    def get_nominal_voltage_cell(self) -> float:
        """Calculated at SOC 50 %."""
        nominal_voltage_cell = 1.39
        return nominal_voltage_cell

    def get_internal_resistance(self, redox_flow_state: RedoxFlowState) -> float:
        power = redox_flow_state.power
        pressure_drop = redox_flow_state.pressure_drop_anolyte
        time_pump = redox_flow_state.time_pump
        power_per_stack = power / (self.get_parallel_scale() * self.get_serial_scale())
        rint_base = 1.8 * self.get_cell_per_stack() / self.get_specific_cell_area()
        if 500 < abs(power_per_stack) < 1000 and pressure_drop == 0 and 0 > time_pump > -450:
            rint_pump = 0.009 + 0.027 * (abs(time_pump) / 450)
        else:
            rint_pump = 0
        rint = rint_base + rint_pump
        self.__log.debug('Resistant stack: ' + str(rint) + ' Ohm, resistant module: ' +
                         str(rint / self.get_parallel_scale() * self.get_serial_scale()))
        return float(rint / self.get_parallel_scale() * self.get_serial_scale())

    def get_cell_per_stack(self) -> int:
        return self.__CELL_NUMBER

    def get_min_voltage(self) -> float:
        return self.__MIN_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_max_voltage(self) -> float:
        return self.__MAX_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_self_discharge_current(self, redox_flow_state: RedoxFlowState) -> float:
        soc_stack = redox_flow_state.soc_stack
        current = redox_flow_state.current / self.get_parallel_scale()
        shunt_current_equivalent_single_stack = float(self.__shunt_interp2d(current, soc_stack))
        shunt_current_equivalent_total = (shunt_current_equivalent_single_stack * 0.81 *
                                          math.exp(0.198 * self.get_serial_scale()))
        self_discharge_current = self.__SELF_DISCHARGE_CURRENT * soc_stack
        self_discharge_current_total = ((shunt_current_equivalent_total + self_discharge_current) *
                                        self.get_cell_per_stack() * self.get_serial_scale() * self.get_parallel_scale())
        return self_discharge_current_total

    def get_specific_cell_area(self) -> float:
        return self.__CELL_AREA

    def get_electrode_thickness(self) -> float:
        return self.__ELECTRODE_THICKNESS / 100

    def get_electrode_porosity(self) -> float:
        return self.__ELECTRODE_POROSITY

    def get_hydraulic_resistance(self) -> float:
        return self.__hydraulic_resistance

    def dependent_parameters(self) -> bool:
        return self.__DEPENDENT_PARAMETERS

    def close(self):
        super().close()
        self.__log.close()



from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.technology.hydrogen.fuel_cell.thermal.thermal_model import ThermalModel


class SimpleThermalModel(ThermalModel):
    """
    Simple Thermal Model with assuming a direct dependency to the fuel cell current based on information of Ballard's
    "Product Manual and Integration Guide Integrating the 1020ACS into a System", 2016
    """

    def __init__(self, temperature: float):
        super().__init__()
        self.__temperature = temperature - 273.15  # K -> C

    def calculate(self, time, state: FuelCellState) -> None:
        self.__temperature = -0.53 * state.current + 26.01

    def get_temperature(self) -> float:
        return self.__temperature

    def get_pump_power(self) -> float:
        return 0.0

    def get_water_flow_stack(self) -> float:
        return 0.0

    def get_power_water_heating(self) -> float:
        return 0.0



import os
import sys
import threading
import time
from multiprocessing import Queue
from queue import Empty

from .utils.utilities import remove_file


class ConsolePrinter(threading.Thread):

    """
    ConsolePrinter is a thread providing information of the current status of all concurrent simulations.
    """

    UPDATE: float = 1
    CUT_OFF_LENGTH: int = 120
    SPACES: int = 3
    FILE: str = os.getcwd() + '/' + 'progress.log'
    STOP_SIGNAL: str = 'STOP'
    REGISTER_SIGNAL: str = 'REGISTER'

    def __init__(self, queue: Queue):
        """
        Constructor of ConsolePrinter

        Parameters
        ----------
        queue :
            Multiprocessing queue exchanging information between concurrent processes

        """
        super().__init__()
        self.__output: dict = dict()
        self.__running: bool = True
        self.__registered: dict = dict()
        self.__file_writer = None
        self.__queue: Queue = queue
        remove_file(self.FILE)

    def register(self, name: str) -> None:
        """
        Register process with name

        Parameters
        ----------
        name :
            process name

        Returns
        -------

        """
        if name not in self.__registered.keys():
            self.__registered[name] = True
        # if self.not_started_yet:
            # self.__file_writer = open(self.FILE, mode='a+')
            # self.start()

    @property
    def not_started_yet(self) -> bool:
        return self.__registered.keys() and not self.is_alive() and self.__running

    def set(self, name: str, line: str) -> None:
        """
        Setting value for process name if process is registered

        Parameters
        ----------
        name :
            process name
        line :
            value to display for process

        Returns
        -------

        """
        if name in self.__registered.keys():
            if self.__registered[name]:
                self.__output[name] = line

    def stop(self, name) -> None:
        """
        Signals that ConsolePrinter is not used for process with 'name' anymore

        Parameters
        ----------
        name :
            name of process

        Returns
        -------

        """
        self.__registered[name] = False
        keys: [str] = list(self.__output.keys())
        if name in keys:
            del self.__output[name]
        stopping: bool = True
        for running in self.__registered.values():
            stopping = stopping and not running
        # if all threads are not running any more stopping stays true -> running = false!
        # This could stop further simulations from running since they have a blocking call for register and stop,
        # only if the first set of simulations finishes before the next simulations are registered.
        # Todo: Fix this! (MM)
        # self.__running = not stopping

    def stop_immediately(self) -> None:
        """
        Stops ConsolePrinter immediately for all (!) processes
        Returns
        -------

        """
        self.__running = False

    @property
    def max_key_length(self) -> int:
        """

        Returns
        -------
        int:
            maximal key length of all processes
        """
        max_length: int = 0
        for key in self.__output.keys():
            length: int = len(key)
            if length > max_length:
                max_length = length
        return max_length

    # Moving cursor and delete multiple lines does not work on windows. Hence, string gets cut off.
    # @staticmethod
    # def move_up(lines: str) -> None:
    #     for _ in range(lines.count('\n')):
    #         sys.stdout.write("\033[F")  # move cursor to the beginning of the previous line
    #         # sys.stdout.write('\b' * len(lines))  # Move back cursor
    #         # sys.stdout.write('\033[A')  # move cursor up one line
    #         # sys.stdout.write('\033[K')  # Clear line
    #         # sys.stdout.write('\x1b[A')  # Cursor up one line
    #         # sys.stdout.write('\x1b[1A\x1b[2K')
    #         # sys.stdout.flush()

    def cut_off(self, line: str) -> str:
        if len(line) > self.CUT_OFF_LENGTH:
            return line[:self.CUT_OFF_LENGTH] + ' ...'
        else:
            return line

    @property
    def line(self) -> str:
        line: str = '\r' # '\r'
        string_format: str = '{:'+str(self.max_key_length)+'s}'
        for key in self.__output.keys():
            line += '[' + string_format.format(key) + ': ' + self.__output[key] + ']' + ' ' * self.SPACES
        return self.cut_off(line) + ''

    @staticmethod
    def clear_screen():
        # os.system('cls' if os.name == 'nt' else 'clear')
        print('\n' * 10000)

    def update(self) -> None:
        try:
            while not self.__queue.empty():
                output: dict = self.__queue.get_nowait()
                items = output.items()
                for key, value in items:
                    if value == self.REGISTER_SIGNAL:
                        self.register(key)
                    elif value == self.STOP_SIGNAL:
                        self.stop(key)
                    else:
                        self.set(key, value)
        except Empty:
            return

    def run(self) -> None:
        while self.__running:
            # self.clear_screen()
            self.update()
            line: str = self.line
            # self.__file_writer.write(line)
            # self.__file_writer.flush()
            sys.stdout.write(line)
            sys.stdout.flush()
            # print(self.__queue.qsize())
            time.sleep(self.UPDATE)
        # self.__file_writer.close()



import pandas as pd

from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig
from simses.commons.constants import Hydrogen
from simses.commons.log import Logger
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel


class PemElectrolyzer(ElectrolyzerStackModel):
    """An PEM-Electrolyzer is a special typ of an Electrolyzer"""

    __POWER_HEADER = 'cellpower at 1 bar'
    __VOLTAGE_IDX = 1
    __POWER_IDX = 1
    __CURRENT_IDX = 0
    __GEOM_AREA_STACK = 50  # cm2
    __POWER = 100  # W
    __WATER_POWER_DENSITY = 0.125  # kg / kW
    __SPEC_MAX_POWER: float = 10.8  # W/cm2
    __MAX_CURRENT_DENSITY: float = 5.0  # A/cm2
    __MIN_CURRENT_DENSITY: float = 0.0 # mA
    __HEAT_CAPACITY: float = 4.933  # J/K/W from: Operational experience and Control strategies for a stand-alone power system based on renewable energy and hydrogen

    __part_pressure_h2: float = 0.526233372104
    __part_pressure_o2: float = 0.526233372104
    __sat_pressure_h2o: float = 0.466925594058

    def __init__(self, electrolyzer_maximal_power: float, electrolyzer_data_config: ElectrolyzerDataConfig):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__NOM_STACK_POWER: float = electrolyzer_maximal_power
        self.__MAX_STACK_POWER: float = electrolyzer_maximal_power
        self.__GEOM_AREA_STACK: float = electrolyzer_maximal_power / self.__SPEC_MAX_POWER
        self.__HEAT_CAPACITY_STACK: float = self.__HEAT_CAPACITY * self.__NOM_STACK_POWER
        self.__NUMBER_CELLS: float = 1.0
        self.__PC_FILE = electrolyzer_data_config.pem_electrolyzer_pc_file
        self.__POWER_FILE = electrolyzer_data_config.pem_electrolyzer_power_file
        self.__polarization_curve = pd.read_csv(self.__PC_FILE)  # V/(mA/cm2)
        self.__power_curve = pd.read_csv(self.__POWER_FILE)  # W/cm2
        self.__current_arr = self.__polarization_curve.iloc[:, self.__CURRENT_IDX] * self.__GEOM_AREA_STACK  # mA
        self.__polarization_arr = self.__polarization_curve.iloc[:, self.__VOLTAGE_IDX]  # V
        self.__power_arr = self.__power_curve.iloc[:, self.__POWER_IDX] * self.__GEOM_AREA_STACK  # W
        self.__voltage: float = 0.0
        self.__current: float = 0.0
        self.__hydrogen_generation: float = 0.0

    def calculate(self, power: float, state: ElectrolyzerState):
        power_idx = abs(self.__power_curve[self.__POWER_HEADER] * self.__GEOM_AREA_STACK - power).idxmin()
        self.__voltage = self.__polarization_curve.iloc[power_idx, self.__VOLTAGE_IDX]  # V
        self.__current = self.__power_curve.iloc[power_idx, self.__CURRENT_IDX] * self.__GEOM_AREA_STACK  # mA
        self.__hydrogen_generation = self.__current / (2 * Hydrogen.FARADAY_CONST)  # mol/s

    def get_current(self):
        # print('the current is: ' + str(self.__current) + ' mA')
        return self.__current

    def get_hydrogen_production(self):
        # print('the massflow of hydrogen ist: ' + str(self.__hydrogen_generation) + ' mol/s')
        return self.__hydrogen_generation

    def get_oxygen_production(self):
        return self.__hydrogen_generation / 2

    def get_voltage(self):
        return self.__voltage

    def get_water_use(self):
        return self.__hydrogen_generation  # water use is the same as hydrogen generation

    def get_number_cells(self):
        return self.__NUMBER_CELLS

    def get_nominal_stack_power(self):
        return self.__NOM_STACK_POWER

    def get_heat_capacity_stack(self):
        return self.__HEAT_CAPACITY_STACK

    def get_partial_pressure_h2(self):
        return self.__part_pressure_h2

    def get_partial_pressure_o2(self):
        return self.__part_pressure_o2

    def get_sat_pressure_h2o(self):
        return self.__sat_pressure_h2o

    def get_geom_area_stack(self):
        return self.__GEOM_AREA_STACK

    def get_reference_voltage_eol(self, resistance_increase: float, exchange_currentdensity_decrease: float) -> float:
        return 1.0

    def get_current_density(self) -> float:
        return self.__current / self.get_geom_area_stack()

    def get_water_in_stack(self) -> float:
        return self.__WATER_POWER_DENSITY * self.get_nominal_stack_power() / 1000.0

    def get_nominal_current_density(self) -> float:
        return 2.0

    def close(self):
        self.__log.close()

    def get_thermal_resistance_stack(self) -> float:
        pass



from abc import ABC, abstractmethod

from simses.commons.state.technology.redox_flow import RedoxFlowState


class CapacityDegradationModel(ABC):
    """Model for the capacity degradation effects of a redox flow battery, which can not be reversed by electrolyte
    remixing."""
    def __init__(self, capacity: float):
        super().__init__()
        self.__capacity_start = capacity

    def update(self, time: float, redox_flow_state: RedoxFlowState) -> None:
        """
        Calculates the capacity degradation and updates the value as well as the state-of-health (SOH).

        Parameters
        ----------
        time : float
            Current simulation time in s.
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------

        """
        redox_flow_state.capacity -= self.get_capacity_degradation(time, redox_flow_state)  # Wh
        redox_flow_state.soh = redox_flow_state.capacity / self.__capacity_start

    @abstractmethod
    def get_capacity_degradation(self, time: float, redox_flow_state: RedoxFlowState) -> float:
        """
        Determination of the capacity degradation.

        Parameters
        ----------
        time : float
            Current simulation time in s.
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float :
            Capacity degradation per time step in Wh.
        """
        pass

    @abstractmethod
    def close(self):
        """Closing all resources in CapacityDegradationModel."""
        pass



from datetime import datetime

import numpy
from pytz import timezone

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.abstract_revenue_stream import RevenueStream
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.commons.config.analysis.economic import EconomicAnalysisConfig


class ElectricityConsumptionRevenueStream(RevenueStream):

    """" Calculates the yearly costs for electrictiy consumption for electrolyzer"""

    __UTC: timezone = timezone('UTC')
    __BERLIN: timezone = timezone('Europe/Berlin')

    def __init__(self, energy_management_data: EnergyManagementData, system_data: SystemData, economic_analysis_config: EconomicAnalysisConfig):
        super().__init__(energy_management_data, system_data, economic_analysis_config)
        self.__electricity_price_grid = economic_analysis_config.electricity_price
        self.__electricity_price_renewable = economic_analysis_config.renewable_electricity_price
        self.__cashflow_list = []
        self.__grid_cost_list = []
        self.__renewable_cost_list = []
        self.__Ws_to_kWh = 1/1000 * 1/3600
        self.__shorter_than_one_year = True

    def get_cashflow(self) -> numpy.ndarray:
        time: numpy.ndarray = self._energy_management_data.time
        timestep = time[1] - time[0]
        total_power_system: numpy.ndarray = self._system_data.power
        renewable_power_total: numpy.ndarray = self._energy_management_data.pv_power
        grid_power = total_power_system - renewable_power_total
        renewable_power_used = numpy.add(renewable_power_total, grid_power.clip(max=0))

        cost_renewable_energy_per_timestep = numpy.asarray([power * self.__electricity_price_renewable if power >= 0
                                                            else 0 for power in renewable_power_used]) * timestep * self.__Ws_to_kWh
        cost_grid_power_per_timestep = numpy.asarray([power * self.__electricity_price_grid if power >= 0 else 0 for
                                                      power in grid_power]) * timestep * self.__Ws_to_kWh

        # divide into list for each project year
        self.__grid_cost_list = self.cash_time_series_to_project_years(cost_grid_power_per_timestep, time)
        self.__grid_cost_list = [-1 * x for x in self.__grid_cost_list]

        self.__renewable_cost_list = self.cash_time_series_to_project_years(cost_renewable_energy_per_timestep,time)
        self.__renewable_cost_list = [-1*x for x in self.__renewable_cost_list]

        self.__cashflow_list = []
        for g, r in zip(self.__grid_cost_list, self.__renewable_cost_list):
            self.__cashflow_list.append(g+r)
        return numpy.array(self.__cashflow_list)

    def __get_next_billing_year(self, date: datetime) -> datetime:
        """Returns begin of following year 20xx-01-01 00:00:00"""
        return date.replace(year=date.year+1, month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

    def get_evaluation_results(self) -> [EvaluationResult]:
        key_results: [EvaluationResult] = list()
        key_results.append(EvaluationResult(Description.Economical.ElectricityConsumption.ELECTRICITY_COST_GRID,
                                            Unit.EURO, sum(self.__grid_cost_list)))
        key_results.append(EvaluationResult(Description.Economical.ElectricityConsumption.ELECTRICITY_COST_RENEWABLE,
                                            Unit.EURO, sum(self.__renewable_cost_list)))
        key_results.append(EvaluationResult(Description.Economical.ElectricityConsumption.TOTAL_ELECTRICITY_COST,
                                            Unit.EURO, sum(self.__cashflow_list)))
        return key_results

    def get_assumptions(self) -> [EvaluationResult]:
        assumptions: [EvaluationResult] = list()
        assumptions.append(EvaluationResult(Description.Economical.ElectricityConsumption.ELECTRICITY_COST_RENEWABLE,
                                            Unit.EURO_PER_KWH, self.__electricity_price_renewable))
        assumptions.append(EvaluationResult(Description.Economical.ElectricityConsumption.ELECTRICITY_COST_GRID,
                                            Unit.EURO_PER_KWH, self.__electricity_price_grid))
        return assumptions


    def close(self):
        pass




class OperationPriority:
    """Priority options of a operation strategy in a multiuse case
        VERY_HIGH = 0
        HIGH = 1
        MEDIUM = 2
        LOW = 3
        VERY_LOW = 4"""
    VERY_HIGH = 0
    HIGH = 1
    MEDIUM = 2
    LOW = 3
    VERY_LOW = 4



import scipy.integrate as integrate
import math

from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel

# Source:
# Thomas Nemeth, Philipp Schröer, Matthias Kuipers, Dirk Uwe Sauer:
# Lithium titanate oxide battery cells for high-power automotive applications -
# Electro-thermal  properties, aging behavior and cost considerations,
# Journal of Energy Storage 31 (2020) 101656, https://doi.org/10.1016/j.est.2020.101656

class LTOLMOCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector):
        super().__init__(cell_type, cycle_detector)
        self.__log: Logger = Logger(type(self).__name__)

        self.__capacity_loss = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()
        self.__capacity_loss_cyclic = cell_type.get_cyclic_capacity_loss_start()

        self.__A_QLOSS = -1.94 * 10 ** (-5)  # constant
        self.__B_QLOSS = 0.00551  # constant

        self.__resistance_increase = 0
        self.__rinc_cyclic = 0

        self.__A_RINC = 0.0001635  # constant
        self.__B_RINC = -0.04751  # constant

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        qloss: float = self.__capacity_loss_cyclic  # in pu
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()  # in pu
        temperature: float = battery_state.temperature  # in K
        # No data available for Temperatures under 298.15 K and 318.15 K
        # Therefore if Temperature is under/over 298.15 K/318.15 K the simulation uses 298.15 K/318.15 K

        if temperature < 298.15:
            temperature = 298.15
            self.__log.warn('Temperature is under 298.15K but the simulation used 298.15K.')
        if temperature > 318.15:
            temperature = 318.15
            self.__log.warn('Temperature is over 318.15K but the simulation used 318.15K.')

        virtual_fec: float = ((1-qloss) * 100 - 100) / (self.__A_QLOSS * temperature + self.__B_QLOSS)
        rel_capacity_status = lambda fec: ((self.__A_QLOSS * temperature + self.__B_QLOSS)*fec+100) / 100
        fec: float = virtual_fec + delta_fec
        capacity_loss = (1 - rel_capacity_status(fec)) - self.__capacity_loss_cyclic
        # Total cyclic capacity loss
        self.__capacity_loss_cyclic += capacity_loss  # in p.u.
        # Delta cyclic capacity loss
        self.__capacity_loss = capacity_loss * self._cell.get_nominal_capacity()  # in Ah

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        rinc: float = self.__rinc_cyclic
        delta_fec: float = self._cycle_detector.get_delta_full_equivalent_cycle()  # in pu
        temperature: float = battery_state.temperature  # in K

        if temperature < 298.15:
            temperature = 298.15
            self.__log.warn('Temperature is under 298.15K but the simulation used 298.15K.')
        if temperature > 318.15:
            temperature = 318.15
            self.__log.warn('Temperature is over 318.15K but the simulation used 318.15K.')

        virtual_fec: float = (rinc * 100) / (self.__A_RINC * temperature + self.__B_RINC)
        rel_resistance_increase = lambda fec: ((self.__A_RINC * temperature + self.__B_RINC) * fec) / 100
        fec = virtual_fec + delta_fec
        resistance_increase = (rel_resistance_increase(fec)) - self.__rinc_cyclic
        self.__rinc_cyclic += resistance_increase
        self.__resistance_increase = resistance_increase

    def get_degradation(self) -> float:
        capacity_loss = self.__capacity_loss
        self.__capacity_loss = 0    # Set value to 0, because cyclic losses are not calculated in each step
        return capacity_loss

    def get_resistance_increase(self) -> float:
        resistance_increase = self.__resistance_increase
        self.__resistance_increase = 0  # Set value to 0, because cyclic losses are not calculated in each step
        return resistance_increase


    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        self.__log.close()



from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.degradation.capacity_degradation import CapacityDegradationModel
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class ConstHydrogenCurrent(CapacityDegradationModel):
    """Simplified model that considers the reduction in capacity due to a constant hydrogen current for a redox flow
    battery."""

    def __init__(self, capacity: float, stack_module: StackModule):
        super().__init__(capacity)
        self.__log: Logger = Logger(type(self).__name__)
        self.__stack_module = stack_module
        """
        In the model a hydrogen evolution current density of 5 * 10^-8 A/cm^2 is assumed. This creates a capacity 
        degradation under 1 % per year (depends on the energy to power ratio of the storage). Whitehead et al. 
        observed that for a commercial system with suitable operation parameters a capacity degradation under 1 % can 
        be achieved without additional measures like additives in the electrolyte or rebalancing cells.
        
        The hydrogen evolution data from Schweiss et al. of single cell measurements show a hydrogen evolution current 
        density around 5 * 10^-6 A/cm^2 (calculated from the hydrogen evolution rate). Values as high as this 
        overestimate the expected capacity degradation. 
        
        Literature sources: 
        Whitehead, Adam H., and Martin Harrer. "Investigation of a method to hinder charge imbalance in the vanadium 
        redox flow battery." Journal of power sources 230 (2013): 271-276.
        
        Schweiss, Ruediger, Alexander Pritzl, and Christian Meiser. "Parasitic hydrogen evolution at different carbon 
        fiber electrodes in vanadium redox flow batteries." Journal of the Electrochemical Society 163.9 (2016): A2089.        
        """
        self.__hydrogen_current_density = 5 * 10 ** -8  # A/cm^2
        self.__hydrogen_current = (self.__hydrogen_current_density * self.__stack_module.get_specific_cell_area())
        self.__log.debug('The hydrogen generation current (per cell) is: ' + str(self.__hydrogen_current))

    def get_capacity_degradation(self, time: float, redox_flow_state: RedoxFlowState):
        time_passed = time - redox_flow_state.time
        capacity_degradation = (self.__hydrogen_current * time_passed * self.__stack_module.get_nominal_voltage_cell() /
                                3600 * self.__stack_module.get_cell_per_stack() * self.__stack_module.get_serial_scale()
                                * self.__stack_module.get_parallel_scale())
        return capacity_degradation  # Wh

    def close(self):
        super().close()
        self.__log.close()
        self.__stack_module.close()



import scipy.interpolate

from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem


class VanadiumSystem(ElectrolyteSystem):
    """The parameters of VanadiumSystem are based on experimental data of an electrolyte consisting of 1.6 M Vanadium in
    aqueous sulphuric acid (2 M H2SO4) from GfE (Gesellschaft für Elektrometallurgie mbH)."""

    MAX_ELECTROLYTE_TEMPERATURE = 40 + 273.15  # K
    MIN_ELECTROLYTE_TEMPERATURE = 10 + 273.15  # K

    def __init__(self, capacity: float, redox_flow_config: RedoxFlowConfig):
        super().__init__(capacity)
        self.__max_soc = redox_flow_config.max_soc
        self.__concentration_v = 1600  # mol/m^3
        self.__density_electrolyte = 1334  # kg/m^3 mean value from Lisa Hoffmann master theses (2018)
        self.__capacity_density = self.FARADAY * self.__concentration_v / 2
        soc_arr_vis = [0.15, 0.5, 0.85]
        temp_arr_vis = [288.15, 298.15, 308.15]
        visc_arr_an = [[0.00602, 0.00444, 0.00348],
                       [0.00576, 0.00418, 0.00321],
                       [0.00529, 0.00372, 0.00275]]
        self.__visc_interp_an = scipy.interpolate.interp2d(soc_arr_vis, temp_arr_vis, visc_arr_an, kind='linear')
        soc_arr_vis = [0.15, 0.5, 0.85]
        temp_arr_vis = [288.15, 298.15, 308.15]
        visc_arr_ca = [[0.00532, 0.00398, 0.00315],
                       [0.00513, 0.00379, 0.00296],
                       [0.00468, 0.00335, 0.00252]]
        self.__visc_interp_ca = scipy.interpolate.interp2d(soc_arr_vis, temp_arr_vis, visc_arr_ca, kind='linear')
        self.__max_temperature = self.MAX_ELECTROLYTE_TEMPERATURE
        self.__min_temperature = self.MIN_ELECTROLYTE_TEMPERATURE
        self.__min_viscosity = float(self.__visc_interp_ca(self.__max_soc, self.get_min_temperature()))
        self.__max_viscosity = float(self.__visc_interp_an(0.0, self.get_max_temperature()))

    def get_viscosity_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        The parameter for the viscosity are based on experimental measurements performed at ZAE Bayern by Lisa Hoffmann.
        Literature source: Hoffmann, Lisa. Physical properties of a VRFB-electrolyte and their impact on the
        cell-performance. master theses. RWTH Aachen, 2018.
        The temperature dependency at SOC 50 % was extrapolated to the other SOC values.
        """
        soc = redox_flow_state.soc
        temp = redox_flow_state.temperature
        viscosity = self.__visc_interp_an(soc, temp)
        return float(viscosity)

    def get_viscosity_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        The parameter for the viscosity are based on experimental measurements performed at ZAE Bayern by Lisa Hoffmann.
        Literature source: Hoffmann, Lisa. Physical properties of a VRFB-electrolyte and their impact on the
        cell-performance. master theses. RWTH Aachen, 2018.
        The temperature dependency at SOC 50 % was extrapolated to the other SOC values.
        """
        soc = redox_flow_state.soc
        temp = redox_flow_state.temperature
        viscosity = self.__visc_interp_ca(soc, temp)
        return float(viscosity)

    def get_min_viscosity(self):
        return self.__min_viscosity

    def get_max_viscosity(self):
        return self.__max_viscosity

    def get_max_temperature(self) -> float:
        return self.__max_temperature

    def get_min_temperature(self) -> float:
        return self.__min_temperature

    def get_vanadium_concentration(self) -> float:
        return self.__concentration_v  # mol/m^3

    def get_capacity_density(self) -> float:
        return self.__capacity_density  # As/m^3

    def get_electrolyte_density(self) -> float:
        return self.__density_electrolyte

    def close(self):
        super().close()



from simses.commons.profile.file import FileProfile
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.profile.technical.technical import TechnicalProfile


class SocProfile(TechnicalProfile):

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig):
        super().__init__()
        self.__file: FileProfile = FileProfile(config, profile_config.soc_file, delimiter=',',
                                               value_index=profile_config.soc_file_value)

    def next(self, time: float) -> float:
        return self.__file.next(time)

    def close(self):
        self.__file.close()



from configparser import ConfigParser
from os.path import dirname, basename

from simses.analysis.evaluation.abstract_evaluation import Evaluation
from simses.analysis.evaluation.merger import EvaluationMerger
from simses.analysis.factory import AnalysisFactory
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.log import Logger


class StorageAnalysis:

    """
    StorageAnalysis conducts the analysis of the simulated storage systems by SimSES. For each storage technology as
    well as for each (sub)system key performance indicators (KPI) are generated and time series are plotted. All
    information is merged into a HTML file which opens in the standard browser after analysis is finished. The analysis
    is configured by analysis.ini in the config package. Additionally, KPIs for comparison between multiple simulations
    are written to file in batch folder located in results path.
    """

    def __init__(self, path: str, config: ConfigParser, batch_dir: str, version: str):
        """
        Constructor of StorageAnalysis

        Parameters
        ----------
        path :
            path to simulation result folder
        config :
            Optional configs taken into account overwriting values from provided config file
        batch_dir :
            Absolute path of result folder for comparing multiple simulation results
        version :
            current simses version
        """
        self.__path: str = path
        self.__version: str = version
        self.__batch_path: str = batch_dir  # str(Path(self.__path).parent).replace('\\','/') + '/' + BATCH_DIR
        self.__simulation_name: str = basename(dirname(self.__path))
        self.__log: Logger = Logger(__name__)
        self.__config: GeneralAnalysisConfig = GeneralAnalysisConfig(config)
        self.__analysis_config: ConfigParser = config

    def run(self) -> None:
        """
        Executes analysis for all technologies and systems

        Returns
        -------

        """
        result_path: str = self.__config.get_result_for(self.__path)
        factory: AnalysisFactory = AnalysisFactory(result_path, self.__analysis_config, self.__version)
        evaluations: [Evaluation] = factory.create_evaluations()
        evaluation_merger: EvaluationMerger = factory.create_evaluation_merger()
        files_to_transpose: [str] = list()
        self.__log.info('Entering analysis')
        for evaluation in evaluations:
            self.__log.info('Running evaluation ' + type(evaluation).__name__)
            evaluation.run()
            evaluation.write_to_csv(result_path)
            evaluation.write_to_batch(path=self.__batch_path, name=self.__simulation_name,
                                      run=basename(dirname(result_path)))
            files_to_transpose.extend(evaluation.get_files_to_transpose())
            evaluation.close()
        Evaluation.transpose_files(files_to_transpose)
        self.__config.write_config_to(result_path)
        evaluation_merger.merge(evaluations)
        factory.close()
        self.close()

    def close(self) -> None:
        """
        Closing all resources in analysis
        Returns
        -------

        """
        self.__log.close()



from simses.commons.profile.power.power_profile import PowerProfile


class AlternatePowerProfile(PowerProfile):

    """
    AlternatePowerProfile produces an alternating power profile, especially useful for clear defined cycle tests.
    """

    def __init__(self, power_on: float = 1500.0, power_off: float = 0.0, scaling_factor: float = 1.0,
                 time_on: float = 1.0, time_off: float = 1.0):
        """
        Constructor of AlternatePowerProfile

        Parameters
        ----------
        power_on :
            power in W for ON phase, default: 1500.0
        power_off :
            power in W for OFF phase, default: 0.0
        scaling_factor :
            linear scaling of values, default: 1.0
        time_on :
            duration of ON phase in h, default: 1.0
        time_off :
            duration of OFF phase in h, default: 1.0
        """
        super().__init__()
        self.__power_on: float = power_on  # W
        self.__power_off: float = power_off  # W
        self.__scaling_factor: float = scaling_factor
        self.__time_on: float = time_on * 3600.0  # s
        self.__time_off: float = time_off * 3600.0  # s
        self.__last_time: float = 0.0
        self.__active: bool = False

    def __is_active(self, time: float) -> bool:
        time_elapsed: float = time - self.__last_time
        if self.__active:
            if time_elapsed >= self.__time_on:
                self.__active = False
                self.__last_time = time
        else:
            if time_elapsed >= self.__time_off:
                self.__active = True
                self.__last_time = time
        return self.__active

    def next(self, time: float) -> float:
        power = self.__power_on if self.__is_active(time) else self.__power_off
        return power * self.__scaling_factor

    def close(self) -> None:
        pass



from abc import ABC, abstractmethod

from simses.system.auxiliary.auxiliary import Auxiliary


class Pump(Auxiliary, ABC):
    """Pump is an auxiliary. It calculates the necessary pump power in W depending on the pressure losses."""
    def __init__(self):
        super().__init__()

    def ac_operation_losses(self) -> float:
        """
        Calculates the ac operation losses.

        Parameters
        ----------

        Returns
        -------
        float:
            Power in W.
        """
        return self.get_pump_power()

    @abstractmethod
    def calculate_pump_power(self, pressure_loss: float) -> None:
        """
        Calculates the pump power in W.

        Parameters
        ----------
        pressure_loss : float
            Current pressure loss in W. It corresponds to the volume flow times the pressure drop.

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_pump_power(self) -> float:
        """
        Gets pump power in W.

        Returns
        -------
        float:
            Pump power in W.
        """
        pass

    @abstractmethod
    def set_eta_pump(self, flow_rate: float, flow_rate_max: float, flow_rate_min: float) -> None:
        """
        Sets pump efficiency.

        Parameters
        ----------
        flow_rate : float
            Current flow rate.
        flow_rate_max : float
            Maximal needed flow rate of the hydraulic system.
        flow_rate_min : float
            Minimal needed flow rate of the hydraulic system.

        Returns
        -------

        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Closing all resources in Pump."""
        super().close()



import math

import numpy as np
import pandas as pd
from scipy import interpolate
from scipy.interpolate import RegularGridInterpolator

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.prismatic import PrismaticCell
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class Samsung94AhNMCHybrid(CellType):
    """This hybrid cell class uses a temperature dependent OCV curves as well as SOC, temperature and
    charge/discharge-rate dependent internal resistance from lab tests, combined with a published degradation
    model based on Naumann et al. Note: While the cell parameters are from a NMC cell, the degradation model is from
    a LFP cell."""

    __SOC_HEADER = 'SOC'
    __SOC_IDX = 0
    __TEMP_IDX = 1
    __C_Rate_IDX = 2

    __CELL_VOLTAGE = 3.68  # V
    __CELL_CAPACITY = 94.0  # Ah
    __MAX_VOLTAGE: float = 4.15  # V
    __MIN_VOLTAGE: float = 2.7  # V
    __MIN_TEMPERATURE: float = 233.15  # K
    __MAX_TEMPERATURE: float = 333.15  # K
    __MAX_CHARGE_RATE: float = 2.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 2.0  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
    __MASS: float = 2.1  # kg per cell
    __SPECIFIC_HEAT: float = 1000  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K

    __HEIGHT: float = 125.0  # mm
    __WIDTH: float = 45.0  # mm
    __LENGTH: float = 173.0  # mm

    __COULOMB_EFFICIENCY: float = 1.0  # p.u

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE, __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = PrismaticCell(__HEIGHT, __WIDTH, __LENGTH)

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)
        rint_file: str = battery_data_config.nmc_samsung94test_rint_file
        # # Reading out the Ri.csv data
        internal_resistance = pd.read_csv(rint_file)  # Ohm
        soc_arr = internal_resistance.iloc[:, self.__SOC_IDX]
        temp_arr = internal_resistance.iloc[:4, self.__TEMP_IDX]
        c_rate_arr = internal_resistance.iloc[:6, self.__C_Rate_IDX]
        rint_mat_ch = internal_resistance.iloc[:, 3:27]
        rint_mat_dch = internal_resistance.iloc[:, 27:]
        # Converting rint_mat_ch & rint_mat_dch into numpy arrays
        rint_mat_ch = rint_mat_ch.values
        rint_mat_dch = rint_mat_dch.values
        # Initializing empty 3D arrays for Rint_ch, Rint_dch - shape - (6, 21, 4)
        rint_mat_ch_tensor = np.ones((6, 21, 4))
        rint_mat_dch_tensor = np.ones((6, 21, 4))
        cursor = 0
        # Fill both tensors - rint_mat_ch_tensor & rint_mat_dch_tensor - with data from excel sheet
        for i in range(4):
            for j in range(6):
                rint_mat_ch_tensor[j, :, i] = rint_mat_ch_tensor[j, :, i] * rint_mat_ch[:, cursor]
                rint_mat_dch_tensor[j, :, i] = rint_mat_dch_tensor[j, :, i] * rint_mat_dch[:, cursor]
                cursor += 1
        self.__rint_ch_rgi = RegularGridInterpolator((c_rate_arr, soc_arr, temp_arr), rint_mat_ch_tensor)  # interpolation with SOC in p.u.
        self.__rint_dch_rgi = RegularGridInterpolator((c_rate_arr, soc_arr, temp_arr), rint_mat_dch_tensor)  # interpolation with SOC in p.u.

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        # temperature window limited between 5°C and 35°C due to test conditions
        if battery_state.temperature > 308.15:
            temperature: float = 308.15
            self.__log.warn("The cell resistance model is only parameterized up to 35°C.")
        elif battery_state.temperature < 278.15:
            temperature: float = 278.15
            self.__log.warn("The cell resistance model is only parameterized down to 5°C.")
        else:
            temperature: float = battery_state.temperature
        # current power limited between 0.25 C-rate and 1.5 C-rate due to test conditions
        c_rate: float = battery_state.current / self.get_nominal_capacity()
        if c_rate < 0.25:
            c_rate = 0.25
            # self.__log.warn("The cell resistance model only works with minimum current power of 0.25 C-rate.")
        if c_rate > 1.5:
            c_rate = 1.5
            self.__log.warn("The cell resistance model only works with maximum current power of 1.5 C-rate.")
        soc: float = max(0.0, min(1.0, battery_state.soc))
        # interpolation phase Ri - for charge and discharge direction
        if battery_state.is_charge:
            rint = self.__rint_ch_rgi([c_rate, soc, temperature]) / 1000  # division with 1000 - change values into Ohm
        else:
            rint = self.__rint_dch_rgi([c_rate, soc, temperature]) / 1000  # division with 1000 - change values into Ohm
        return float(rint) / self.get_parallel_scale() * self.get_serial_scale()

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        soc = battery_state.soc
        # temperature range for measured values
        temp_arr = np.array([278.15, 288.15, 298.15, 308.15])
        # temperature window limited between 5°C and 35°C due to test conditions
        if battery_state.temperature > 308.15:
            temperature: float = 308.15
            self.__log.warn("The cell OCV model is only parameterized up to 35°C.")
        elif battery_state.temperature < 278.15:
            temperature: float = 278.15
            self.__log.warn("The cell OCV model is only parameterized down to 5°C.")
        else:
            temperature: float = battery_state.temperature

        # curve fit parameters from OCV fitting tool
        a1 = np.array([-23.9972, -12.7184, 3.3479, -2.8352])
        a2 = np.array([-84.1166, -64.9682, -6.7241, 1.3448])
        a3 = np.array([-0.8678, -19.9681, 2.5958, 5.5920])
        a4 = np.array([-2.1020, -7.6012, -61.9684, -68.9266])
        b1 = np.array([0.1387, 0.0995, 0.6350, 0.2969])
        b2 = np.array([0.0041, 0.0024, 1.4376, -1.0537])
        k0 = np.array([3.3630, 4.2083, 4.5868, 4.9401])
        k1 = np.array([0.1576, 1.5088, 3.1768, -1.9207])
        k2 = np.array([0.3128, 0.4996, -3.8418, -4.9977])
        k3 = np.array([4.8573, 0.0915, -4.6932, -0.3352])
        k4 = np.array([-3.3540, -3.0541, 0.3618,  0.3535])
        k5 = np.array([0.8500, 0.9240, 0.9949, 0.9982])

        # calculate ocv for measured temperatures
        ocv_all_temps = np.empty(len(a1))
        for i in range(0, len(a1)):
            ocv_all_temps[i] = k0[i] + \
                  k1[i] / (1 + math.exp(a1[i] * (soc - b1[i]))) + \
                  k2[i] / (1 + math.exp(a2[i] * (soc - b2[i]))) + \
                  k3[i] / (1 + math.exp(a3[i] * (soc - 1))) +\
                  k4[i] / (1 + math.exp(a4[i] * soc)) +\
                  k5[i] * soc
        values_interp = interpolate.interp1d(temp_arr, ocv_all_temps)
        ocv = values_interp(temperature)
        return ocv * self.get_serial_scale()

    def get_capacity(self, battery_state: LithiumIonState) -> float:
        return self.get_nominal_capacity()

    def close(self):
        self.__log.close()



from simses.commons.profile.economic.market import MarketProfile


class ConstantPrice(MarketProfile):

    def __init__(self, price: float):
        super(ConstantPrice, self).__init__()
        self.__price: float = price

    def next(self, time: float) -> float:
        return self.__price

    def initialize_profile(self) -> None:
        pass

    def profile_data_to_list(self, sign_factor=1) -> ([float], [float]):
        pass

    def close(self):
        pass



from configparser import ConfigParser

from simses.commons.config.analysis.analysis_config import AnalysisConfig


class EconomicAnalysisConfig(AnalysisConfig):

    """
    Provides
    """

    SECTION: str = 'ECONOMIC_ANALYSIS'

    INVESTMENT_COSTS: str = 'INVESTMENT_COSTS'
    SPECIFIC_INVESTMENT_COSTS_ENERGY: str = 'SPECIFIC_INVESTMENT_COSTS_ENERGY'
    SPECIFIC_INVESTMENT_COSTS_POWER: str = 'SPECIFIC_INVESTMENT_COSTS_POWER'
    USE_SPECIFIC_COSTS: str = 'USE_SPECIFIC_COSTS'
    DISCOUNT_RATE: str = 'DISCOUNT_RATE'
    ELECTRICITY_PRICE: str = 'ELECTRICITY_PRICE'
    PV_FEED_IN_TARIFF: str = 'PV_FEED_IN_TARIFF'
    DEMAND_CHARGE_BILLING_PERIOD: str = 'DEMAND_CHARGE_BILLING_PERIOD'
    DEMAND_CHARGE_PRICE: str = 'DEMAND_CHARGE_PRICE'
    DEMAND_CHARGE_AVERAGE_INTERVAL: str = 'DEMAND_CHARGE_AVERAGE_INTERVAL'
    FCR_PRICE: str = 'FCR_PRICE'
    FCR_USE_PRICE_TIMESERIES: str = 'FCR_USE_PRICE_TIMESERIES'
    IDM_PRICE: str = 'IDM_PRICE'
    IDM_USE_PRICE_TIMESERIES: str = 'IDM_USE_PRICE_TIMESERIES'
    ADD_OPERATION_AND_MAINTENANCE_REVENUE_STREAM: str = 'ADD_OPERATION_AND_MAINTENANCE_REVENUE_STREAM'
    ANNUAL_RELATIVE_OPERATION_AND_MAINTENANCE_COSTS: str = 'ANNUAL_RELATIVE_OPERATION_AND_MAINTENANCE_COSTS'
    RENEWABLE_ELECTRICITY_PRICE: str = 'RENEWABLE_ELECTRICITY_PRICE'

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)

    @property
    def investment_costs(self) -> float:
        """Defines initial investment costs for the energy storage system"""
        return float(self.get_property(self.SECTION, self.INVESTMENT_COSTS))

    @property
    def specific_investment_costs_energy(self) -> float:
        """Defines specific investment costs for the energy storage system in Euro/kWh"""
        return float(self.get_property(self.SECTION, self.SPECIFIC_INVESTMENT_COSTS_ENERGY))

    @property
    def specific_investment_costs_power(self) -> float:
        """Defines additional specific investment costs for the energy storage system in Euro/kW"""
        return float(self.get_property(self.SECTION, self.SPECIFIC_INVESTMENT_COSTS_POWER))

    @property
    def use_specific_costs(self) -> bool:
        """Determine if specific costs or absolute costs are used to determine the investment costs"""
        return self.get_property(self.SECTION, self.USE_SPECIFIC_COSTS) in ['True']

    @property
    def discount_rate(self) -> float:
        """Defines the discount rate in p.u."""
        return float(self.get_property(self.SECTION, self.DISCOUNT_RATE))

    @property
    def electricity_price(self) -> float:
        """Defines the electricity price in Euro/kWh"""
        return float(self.get_property(self.SECTION, self.ELECTRICITY_PRICE))

    @property
    def pv_feed_in_tariff(self) -> float:
        """Defines the feed in tariff in Euro/kWh"""
        return float(self.get_property(self.SECTION, self.PV_FEED_IN_TARIFF))

    @property
    def demand_charge_billing_period(self) -> str:
        """Defines the billing period. Choose 'yearly' or 'monthly'"""
        return self.get_property(self.SECTION, self.DEMAND_CHARGE_BILLING_PERIOD)

    @property
    def demand_charge_price(self) -> float:
        """Defines the demand charge price in Euro/kW"""
        return float(self.get_property(self.SECTION, self.DEMAND_CHARGE_PRICE))

    @property
    def demand_charge_average_interval(self) -> float:
        """Interval length for determining power average in seconds"""
        return float(self.get_property(self.SECTION, self.DEMAND_CHARGE_AVERAGE_INTERVAL))

    @property
    def fcr_price(self) -> float:
        """FCR price in Euro per kW per day"""
        return float(self.get_property(self.SECTION, self.FCR_PRICE))

    @property
    def fcr_use_price_timeseries(self) -> bool:
        """Determine if constant FCR price or prices from FCR_PRICE_PROFILE are used."""
        return self.get_property(self.SECTION, self.FCR_USE_PRICE_TIMESERIES) in ['True']

    @property
    def idm_price(self) -> float:
        """IDM price in Euro per kWh"""
        return float(self.get_property(self.SECTION, self.IDM_PRICE))

    @property
    def idm_use_price_timeseries(self) -> bool:
        """Determine if constant IDM price or prices from IDM_PRICE_PROFILE are used."""
        return self.get_property(self.SECTION, self.IDM_USE_PRICE_TIMESERIES) in ['True']

    @property
    def add_o_and_m_revenue_stream(self) -> bool:
        """ Defines if operation and maintenance revenue stream is used """
        return self.get_property(self.SECTION, self.ADD_OPERATION_AND_MAINTENANCE_REVENUE_STREAM) in ['True']

    @property
    def annual_realative_o_and_m_costs(self) -> float:
        """Relative annual costs for operation and maintenance related to investment costs in p.u."""
        return float(self.get_property(self.SECTION, self.ANNUAL_RELATIVE_OPERATION_AND_MAINTENANCE_COSTS))

    @property
    def renewable_electricity_price(self) -> float:
        """ price for electricity out of renewable source in Euro per kWh """
        return float(self.get_property(self.SECTION, self.RENEWABLE_ELECTRICITY_PRICE))



import math as m

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

__TEMPERATURE_CELL_C: float = 80  # °C
__TEMPERATURE_CELL_K: float = 273.15 + __TEMPERATURE_CELL_C  # K

__IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
__FARADAY_CONST: float = 96485.3321  # As/mol

__CONTACT_RESISTANCE = 25 * 10 ** -3  # Ohm * cm2

__AREA_CELL = 50  # cm2
__ALPHA_ANODE = 0.49
__ALPHA_CATHODE = 0.255
__RUGOSITY_ANODE = 600
__RUGOSITY_CATHODE = 240
__ACT_ENERGY_ANODE = 62836  # J/mol
__ACT_ENERGY_CATHODE = 24359  # J/mol
__K_0_ANODE = 4.63 * 10 ** (-3)  # mol/K/s/m²
__K_0_CATHODE = 1 * 10 ** (-2)  # mol/K/s/m²
__LAMBDA = 22.5  # degree of humidification
__THIKNESS_NAFION = 177.8  # um
__CONDUCTIVITY_NAFION = (0.005139 * __LAMBDA - 0.00326) * m.exp(1268 * (1 / 303 - 1 / __TEMPERATURE_CELL_K))  # S/cm 0.14
__mem_resistance = __THIKNESS_NAFION * 10 ** -4 / (__CONDUCTIVITY_NAFION)  # Ohm * cm2

__ATM = 1.013  # bar

__PRESSURE_ANODE = 1  # bar   total pressure Anode
__PRESSURE_CATHODE = 1  # bar   total pressure Cathode

__pressure_sat_h2o = 10 ** (-2.1794 + 0.02953 * __TEMPERATURE_CELL_C - 9.1837 * 10 ** -5 * __TEMPERATURE_CELL_C ** 2 + 1.4454 * 10 ** -7 *
                            __TEMPERATURE_CELL_C ** 3)

__pressure_h2 = (__PRESSURE_CATHODE - __pressure_sat_h2o) / __ATM  # bar to atm
__pressure_o2 = (__PRESSURE_ANODE - __pressure_sat_h2o) / __ATM  # bar to atm

__open_circuit_voltage = []

__open_circuit_voltage = __IDEAL_GAS_CONST * __TEMPERATURE_CELL_K / (2 * __FARADAY_CONST) * \
                         m.log(__pressure_h2 * __pressure_o2 ** 0.5) + (1.5241 - 1.2261 * 10 ** -3 * __TEMPERATURE_CELL_K +
                                                                        1.1858 * 10 ** -5 * __TEMPERATURE_CELL_K * m.log(
                        __TEMPERATURE_CELL_K) + 5.6692 * 10 ** -7 *
                                                                        __TEMPERATURE_CELL_K ** 2)  # V

__current_density = np.linspace(0, 5000, 500) * 10 ** -3  # mA

__ex_current_dens_anode = __RUGOSITY_ANODE * 2 * __FARADAY_CONST * __K_0_ANODE * __TEMPERATURE_CELL_K * m.exp(- __ACT_ENERGY_ANODE /  (__IDEAL_GAS_CONST * __TEMPERATURE_CELL_K)) / 10000 # A/cm² nach Formel aus schriftl. Ausarbeitung nicht 1737 sondern 893
__ex_current_dens_cathode = __RUGOSITY_CATHODE * 2 * __FARADAY_CONST * __K_0_CATHODE *__TEMPERATURE_CELL_K * m.exp(- __ACT_ENERGY_CATHODE / (__IDEAL_GAS_CONST * __TEMPERATURE_CELL_K)) / 10000  # A/cm²
_cell_voltage = np.ones(len(__current_density))

_spec_cell_power = np.ones(len(__current_density))

for i in range(0, len(__current_density)):
    _cell_voltage[i] = __open_circuit_voltage + __IDEAL_GAS_CONST * __TEMPERATURE_CELL_K / \
                       (2 * __ALPHA_ANODE * __FARADAY_CONST) * m.asinh(__current_density[i] / (__ex_current_dens_anode*2)) + \
                       __IDEAL_GAS_CONST * __TEMPERATURE_CELL_K / \
                       (2 * __ALPHA_CATHODE * __FARADAY_CONST) * m.asinh(__current_density[i] /
                                                                     (__ex_current_dens_cathode*2)) + __CONTACT_RESISTANCE \
                        * __current_density[i] + __mem_resistance * __current_density[i]

for i in range(0, len(__current_density)):
    _spec_cell_power[i] = _cell_voltage[i] * __current_density[i]  # (W/cm2)


x = __current_density  # A/cm2
y1 = _cell_voltage  # V
y2 = _spec_cell_power  # W/cm2

ax1 = plt.subplot(211)
plt.plot(x, y1)

ax2 = plt.subplot(212)
plt.plot(x, y2)
plt.show()

dict = {'currentdensity': __current_density, 'cellvoltage at 1 bar': _cell_voltage}
dict2 = {'currentdenstiy': __current_density, 'cellpower at 1 bar': _spec_cell_power}

df = pd.DataFrame(dict)
df2 = pd.DataFrame(dict2)

df.to_csv('polarizationcurve1bar.csv', index=False)
df2.to_csv('powercurve1bar.csv', index=False)



class ThermalCellProperties:

    """
    ThermalCellProperties inherits all necessary data for describing the thermal behaviour of a cell.
    """

    def __init__(self, min_temperature: float, max_temperature: float, mass: float,
                 specific_heat: float, convection_coeffecient: float):
        """
        Constructor for ThermalCellProperties

        Parameters
        ----------
        min_temperature :
            minimum allowed temperature for the cell in K
        max_temperature :
            maximum allowed temperature for the cell in K
        mass :
            mass of one cell in kg
        specific_heat :
            specific heat of the cell in J/kgK
        convection_coeffecient :
            convection coefficient in W/m2K
        """
        self.__MIN_TEMPERATURE: float = min_temperature
        self.__MAX_TEMPERATURE: float = max_temperature
        self.__MASS: float = mass
        self.__SPECIFIC_HEAT: float = specific_heat
        self.__CONVECTION_COEFFICIENT: float = convection_coeffecient

    def get_min_temperature(self) -> float:
        """

        Returns
        -------
        float:
            minimum allowed temperature for the cell in K
        """
        return self.__MIN_TEMPERATURE

    def get_max_temperature(self) -> float:
        """

        Returns
        -------
        float:
            maximum allowed temperature for the cell in K
        """
        return self.__MAX_TEMPERATURE

    def get_specific_heat(self) -> float:
        """

        Returns
        -------
        float:
            specific heat of the cell in J/kgK
        """
        return self.__SPECIFIC_HEAT

    def get_convection_coefficient(self) -> float:
        """

        Returns
        -------
        float:
            convection coefficient in W/m2K
        """
        return self.__CONVECTION_COEFFICIENT

    def get_mass(self) -> float:
        """

        Returns
        -------
        float:
            mass of one cell in kg
        """
        return self.__MASS



from simses.analysis.data.electrolyzer import ElectrolyzerData
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotly_plotting import PlotlyPlotting
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.analysis.evaluation.technical.technical_evaluation import TechnicalEvaluation
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.state.technology.electrolyzer import ElectrolyzerState


class ElectrolyzerTechnicalEvaluation(TechnicalEvaluation):

    title = 'Electrolyzer results'

    def __init__(self, data: ElectrolyzerData, config: GeneralAnalysisConfig, path: str):
        super().__init__(data, config)
        title_extension: str = ' for system ' + self.get_data().id
        self.title += title_extension
        self.__result_path = path
        # self.__electrolyzer = PemElectrolyzerMultiDimAnalytic(16000, ElectrolyzerDataConfig())

    def evaluate(self):
        # super().evaluate()
        data: ElectrolyzerData = self.get_data()
        self.append_result(EvaluationResult(Description.Technical.H2_PRODUCTION_EFFICIENCY_LHV, Unit.PERCENTAGE, self.h2_production_efficiency_lhv()))
        self.append_result(EvaluationResult(Description.Technical.SOH, Unit.PERCENTAGE, self.soh()))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_H2_REACTION, Unit.KWH, data.total_energy_reaction))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_WATER_HEATING, Unit.KWH, data.total_energy_heating))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_WATER_CIRCULATION, Unit.KWH, data.total_energy_pump))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_H2_DRYING, Unit.KWH, data.total_energy_gas_drying))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_H2_COMPRESSION, Unit.KWH, data.total_energy_compressor))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_H2_LHV, Unit.KWH, data.total_energy_h2_lhv))
        self.append_result(EvaluationResult(Description.Technical.TOTAL_H2_PRODUCTION_KG, Unit.KG, data.total_amount_h2_kg))
        self.append_result(EvaluationResult(Description.Technical.TOTAL_H2_PRODUCTION_NM, Unit.NCM, data.total_amount_h2_nqm))
        self.print_results()

    def plot(self) -> None:
        self.fulfillment_plotting()
        self.current_plotting()
        self.current_dens_plotting()
        self.degradation_plotting()
        self.soh_plotting()
        self.hydrogen_production_plotting()
        self.pressures_plotting()
        self.hydrogen_outflow_plotting()
        #self.pressure_anode_plotting()
        self.temperature_plotting()
        # self.power_water_heating_plotting()
        self.power_auxilliaries_1_plotting()
        self.power_auxilliaries_2_plotting()
        # self.efficiency_plotting()

    def h2_production_efficiency_lhv(self) -> float:
        """
        Calculates the hydrogen production efficiency relative to its lower heating value

        Parameters
        -----------
            data: simulation results

        :return:
            float: h2 production efficiency
        """
        data: ElectrolyzerData = self.get_data()
        return data.total_energy_h2_lhv / (data.total_energy_pump + data.total_energy_compressor +
                                           data.total_energy_gas_drying + data.total_energy_heating +
                                           data.total_energy_reaction) * 100

    def soh(self, index=-1):
        data: ElectrolyzerData = self.get_data()
        return data.soh[index] * 100  #  %

    # def efficieny_curve_electrolyzer(self, index=0):
    #     data: ElectrolyzerData = self.get_data()
    #     # timestep = data.time[1] - data.time[0]
    #     # state_index = int(time / timestep)
    #     # values out of the hydrogen state are used for temperature and pressure
    #     # stack_temperature = data.temperature_el[index] - 273.15  # K -> °C
    #     # p_anode = data.pressure_anode_el[index]
    #     # p_cathode = data.pressure_cathode_el[index]
    #     # fixed values for pressure and temperature, so that the efficiency curves are only changed by the values for
    #     # degradation -> resistance increase and exchange_current_decrease
    #     eff_hydrogen_state = ElectrolyzerState(1,1)
    #     eff_hydrogen_state.temperature_el = 80  # °C
    #     eff_hydrogen_state.pressure_anode_el = 0  # barg
    #     eff_hydrogen_state.pressure_cathode_el = 25  # barg
    #     # eff_hydrogen_state.temperature_el = data.temperature_el[index]
    #     # eff_hydrogen_state.pressure_cathode_el = data.pressure_cathode_el[index]
    #     # eff_hydrogen_state.pressure_anode_el = data.pressure_anode_el[index]
    #     eff_hydrogen_state.resistance_increase_el = data.resistance_increase[index]
    #     eff_hydrogen_state.exchange_current_decrease_el = data.exchange_current_dens_decrease[index]
    #     # efficiency_df = self.__electrolyzer.calculate_efficiency_curves(eff_hydrogen_state)
    #     efficiency_df = self.__electrolyzer.get_efficiency_curve(eff_hydrogen_state)
    #     return efficiency_df

    def current_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=ElectrolyzerState.CURRENT, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = [Axis(data=data.current, label=ElectrolyzerState.CURRENT)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def current_dens_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=ElectrolyzerState.CURRENT_DENSITY, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = [Axis(data=data.current_density, label=ElectrolyzerState.CURRENT_DENSITY)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def hydrogen_production_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=ElectrolyzerState.HYDROGEN_PRODUCTION, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = [Axis(data=data.hydrogen_production, label=ElectrolyzerState.HYDROGEN_PRODUCTION)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def pressures_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title='Pressures Cathode', path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.pressure_cathode, label=ElectrolyzerState.PRESSURE_CATHODE,
                          color=PlotlyPlotting.Color.CATHODE_PINK, linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.part_pressure_h2, label=ElectrolyzerState.PART_PRESSURE_H2,
                          color=PlotlyPlotting.Color.CATHODE_PINK, linestyle=PlotlyPlotting.Linestyle.DASHED))
        yaxis.append(Axis(data=data.pressure_anode, label=ElectrolyzerState.PRESSURE_ANODE,
                          color=PlotlyPlotting.Color.ANODE_GREEN, linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.part_pressure_o2, label=ElectrolyzerState.PART_PRESSURE_O2,
                          color=PlotlyPlotting.Color.ANODE_GREEN, linestyle=PlotlyPlotting.Linestyle.DASHED))
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def temperature_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=ElectrolyzerState.TEMPERATURE, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.temperature, label=ElectrolyzerState.TEMPERATURE,
                          color=PlotlyPlotting.Color.TEMPERATURE_RED))
        yaxis.append(Axis(data=data.convection_heat, label=ElectrolyzerState.CONVECTION_HEAT,
                          color=PlotlyPlotting.Color.HEAT_ORANGE))
        plot.lines(xaxis, yaxis, [1])
        self.extend_figures(plot.get_figures())

    # def power_water_heating_el_plotting(self):
    #     data: HydrogenData = self.get_data()
    #     plot: Plotting = PlotlyPlotting(title='Elektrolyzer Stack Heat', path=self.__result_path)
    #     xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=HydrogenState.TIME)
    #     yaxis: [Axis] = list()
    #     yaxis.append(Axis(data=data.power_water_heating_el, label=HydrogenState.POWER_WATER_HEATING_EL,
    #                       color=PlotlyPlotting.Color.BLUE))
    #     plot.lines(xaxis, yaxis, [2,3])
    #     self.extend_figures(plot.get_figures())

    def power_auxilliaries_1_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title='Auxiliaries 1', path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data.power_water_heating, label=ElectrolyzerState.POWER_WATER_HEATING,
                          color=PlotlyPlotting.Color.AC_POWER_BLUE, linestyle=PlotlyPlotting.Linestyle.DASHED))
        yaxis.append(Axis(data.power_pump, label=ElectrolyzerState.POWER_PUMP,
                          color=PlotlyPlotting.Color.AC_POWER_BLUE,linestyle=PlotlyPlotting.Linestyle.DASHED))
        plot.lines(xaxis, yaxis, [1])
        self.extend_figures(plot.get_figures())

    def power_auxilliaries_2_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title='Auxiliaries 2', path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data.power_compressor, label=ElectrolyzerState.POWER_COMPRESSOR,
                          color=PlotlyPlotting.Color.AC_POWER_BLUE, linestyle= PlotlyPlotting.Linestyle.DASHED))
        yaxis.append(Axis(data.power_gas_drying, label=ElectrolyzerState.POWER_GAS_DRYING,
                          color=PlotlyPlotting.Color.RED, linestyle=PlotlyPlotting.Linestyle.DOTTED))
        plot.lines(xaxis, yaxis, [1])
        self.extend_figures(plot.get_figures())

    def hydrogen_outflow_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title='Hydrogen outflow and total injection', path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.hydrogen_outflow, label=ElectrolyzerState.HYDROGEN_OUTFLOW))
        yaxis.append(Axis(data=data.total_h2_production, label=ElectrolyzerState.TOTAL_HYDROGEN_PRODUCTION,
                          color=PlotlyPlotting.Color.CATHODE_PINK, linestyle=PlotlyPlotting.Linestyle.DOTTED))
        plot.lines(xaxis, yaxis, [1])
        self.extend_figures(plot.get_figures())

    def fulfillment_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=ElectrolyzerState.FULFILLMENT, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = [Axis(data=data.storage_fulfillment, label=ElectrolyzerState.FULFILLMENT)]
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def degradation_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title='Degradation Electrolyzer', path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.resistance_increase, label=ElectrolyzerState.RESISTANCE_INCREASE, color=PlotlyPlotting.Color.RESISTANCE_BLACK, linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.resistance_increase_cyclic, label=ElectrolyzerState.RESISTANCE_INCREASE_CYCLIC, color=PlotlyPlotting.Color.RESISTANCE_BLACK, linestyle=PlotlyPlotting.Linestyle.DASHED))
        yaxis.append(Axis(data=data.resistance_increase_calendar, label=ElectrolyzerState.RESISTANCE_INCREASE_CALENDAR, color=PlotlyPlotting.Color.RESISTANCE_BLACK, linestyle=PlotlyPlotting.Linestyle.DOTTED))
        yaxis.append(Axis(data=data.exchange_current_dens_decrease, label=ElectrolyzerState.EXCHANGE_CURRENT_DENS_DECREASE, color=PlotlyPlotting.Color.CURRENT_CYAN))
        plot.lines(xaxis, yaxis, [3])
        self.extend_figures(plot.get_figures())

    def soh_plotting(self):
        data: ElectrolyzerData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=ElectrolyzerState.SOH, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=ElectrolyzerState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.soh, label=ElectrolyzerState.SOH, color=PlotlyPlotting.Color.SOH_GREEN))
        # yaxis.append(Axis(data=data.ref_voltage_el, label=HydrogenState.REFERENCE_VOLTAGE_EL, color=PlotlyPlotting.Color.YELLOW))
        plot.lines(xaxis, yaxis, [1])
        self.extend_figures(plot.get_figures())

    # def efficiency_plotting(self):
    #     efficiency_data_begin = self.efficieny_curve_electrolyzer(0)
    #     efficiency_data_end = self.efficieny_curve_electrolyzer(-1)
    #     current_dens_array = efficiency_data_begin['current density'].to_numpy()
    #     voltage_efficiency_array_begin = efficiency_data_begin['voltage efficiency'].to_numpy()
    #     voltage_efficiency_array_end = efficiency_data_end['voltage efficiency'].to_numpy()
    #     faraday_efficiency_array_begin = efficiency_data_begin['faraday efficiency'].to_numpy()
    #     faraday_efficiency_array_end = efficiency_data_end['faraday efficiency'].to_numpy()
    #     cell_efficiency_array_begin = efficiency_data_begin['cell efficiency'].to_numpy()
    #     cell_efficiency_array_end = efficiency_data_end['cell efficiency'].to_numpy()
    #
    #     file_name: str = 'electrolyzer_efficiency.csv'
    #     file = self.__result_path + file_name
    #     # with open(file, 'w', newline=' ') as file:
    #     #     writer = csv.writer(file)
    #     dict = {'currentdensity': current_dens_array, 'voltage eff begin': voltage_efficiency_array_begin,
    #             'voltage eff end': voltage_efficiency_array_end, 'faraday eff begin': faraday_efficiency_array_begin,
    #             'faraday eff end': faraday_efficiency_array_end, 'cell eff begin': cell_efficiency_array_begin,
    #             'cell eff end': cell_efficiency_array_end}
    #     df = pd.DataFrame(dict)
    #     df.to_csv(file, index=False)
    #
    #     plot: Plotting = PlotlyPlotting(title='Efficiency', path=self.__result_path)
    #     xaxis: Axis = Axis(data=current_dens_array, label='currentdensity')
    #     yaxis: [Axis] = list()
    #     yaxis.append(Axis(data=cell_efficiency_array_begin, label='cell efficiency begin'))
    #     yaxis.append(Axis(data=cell_efficiency_array_end, label='cell efficiency end', color=PlotlyPlotting.Color.RESISTANCE_BLACK))
    #     #yaxis.append(Axis(data=voltage_efficiency_array, label='voltage efficiency', color=PlotlyPlotting.Color.RED))
    #     #yaxis.append(Axis(data=faraday_efficiency_array, label='faraday efficiency', color=PlotlyPlotting.Color.GREEN))
    #     plot.lines(xaxis, yaxis)
    #     self.extend_figures(plot.get_figures())



import csv
import os
from abc import ABC, abstractmethod
import numpy
import pandas
from simses.analysis.data.abstract_data import Data
from simses.analysis.evaluation.result import EvaluationResult
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.log import Logger
from simses.commons.state.abstract_state import State
from simses.commons.utils.utilities import create_directory_for


class Evaluation(ABC):

    """
    Within the evaluation class the analysis of each system and storage technology is conducted. It provides results
    in form of figures and KPIs. The analysis calculations are done with the help of data object provided which accesses
    the simulation data.
    """

    EXT: str = '.csv'

    def __init__(self, data: Data, config: GeneralAnalysisConfig, do_evaluation: bool):
        self.__log: Logger = Logger(__name__)
        self.__data: Data = data
        self.__do_evaluation: bool = do_evaluation
        self.__do_plotting: bool = config.plotting and do_evaluation
        self._print_to_console: bool = config.print_result_to_console and do_evaluation
        self.__export_analysis_to_csv: bool = config.export_analysis_to_csv and do_evaluation
        self.__export_analysis_to_batch: bool = config.export_analysis_to_batch and do_evaluation
        self.__file_name: str = type(self).__name__ + self.__data.id + self.EXT
        self.__evaluation_results: [EvaluationResult] = list()
        self.__time_series: dict = dict()
        self.__files_to_transpose: [str] = list()
        self.__figures: list = list()

    @property
    def evaluation_results(self) -> [EvaluationResult]:
        return self.__evaluation_results

    def append_result(self, evaluation_result: EvaluationResult) -> None:
        self.__evaluation_results.append(evaluation_result)

    def extend_results(self, evaluation_results: [EvaluationResult]) -> None:
        self.__evaluation_results.extend(evaluation_results)

    def append_figure(self, figure) -> None:
        self.__figures.append(figure)

    def extend_figures(self, figures: list) -> None:
        self.__figures.extend(figures)

    def append_time_series(self, name: str, time_series: numpy.ndarray):
        self.__time_series.update({name: time_series})

    def get_files_to_transpose(self) -> [str]:
        return self.__files_to_transpose

    def get_data(self):
        return self.__data

    def get_figures(self) -> list:
        return self.__figures

    def get_file_name(self) -> str:
        return self.__file_name

    @property
    def get_name(self) -> str:
        return type(self.__data).__name__

    @property
    def should_be_considered(self) -> bool:
        return self.__do_evaluation

    def run(self) -> None:
        if self.__do_evaluation:
            self.evaluate()
            if self.__do_plotting:
                self.plot()

    @abstractmethod
    def evaluate(self) -> None:
        pass

    @abstractmethod
    def plot(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    def print_results(self):
        if not self._print_to_console:
            return
        print('\n[' + str(self.get_name).upper() + ' ANALYSIS]' + ' (System ' + self.__data.id + ')')
        for evaluation_result in self.evaluation_results:
            print(evaluation_result.to_console())

    def write_to_csv(self, path: str) -> None:
        if not self.__export_analysis_to_csv:
            return
        file = path + self.__file_name
        with open(file, 'w', newline='') as file:
            writer = csv.writer(file)
            for evaluation_result in self.evaluation_results:
                writer.writerow(evaluation_result.to_csv())

    def write_to_batch(self, path: str, name: str, run: str) -> None:
        if not self.__export_analysis_to_batch:
            return
        create_directory_for(path)
        for evaluation_result in self.evaluation_results:
            file_name = path + evaluation_result.description + self.EXT
            output: list = [name, run, self.get_name, self.__data.id, evaluation_result.value, evaluation_result.unit]
            self.__append_to_file(file_name, output)
        for ts_name, time_series in self.__time_series.items():
            file_name: str = path + ts_name + self.EXT
            if file_name not in self.__files_to_transpose:
                self.__files_to_transpose.append(file_name)
            if not os.path.exists(file_name):
                output: list = [State.TIME]
                output.extend(self.__data.time.tolist())
                self.__append_to_file(file_name, output)
            id: str = name + '_' + run + '_' + self.get_name + '_' + self.__data.id
            output: list = [id]
            output.extend(time_series.tolist())
            self.__append_to_file(file_name, output)
            # data_frame: pandas.DataFrame = pandas.read_csv(file_name, header=0, index=None)
            # data_frame[column_id] = time_series
            # data_frame.to_csv(file_name, index=None)

    def __append_to_file(self, file_name: str, data: list) -> None:
        try:
            with open(file_name, 'a', newline='') as file:
                writer = csv.writer(file)
                writer.writerow(data)
        except FileNotFoundError:
            self.__log.error(file_name + ' could not be found')

    @classmethod
    def transpose_files(cls, files: [str]) -> None:
        transposed: [str] = list()
        for file in files:
            if file not in transposed and os.path.exists(file):
                pandas.read_csv(file, header=None).T.to_csv(
                    file.replace(cls.EXT, '_T' + cls.EXT), header=False, index=False)
                transposed.append(file)



import math as m

from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel


class CalendarDegradationPemElMultiDimAnalyitic(CalendarDegradationModel):
    """
    Calendaric degradation model for PemElectrolyzerMultiDimAnalyitc
    decreases the exchange current density in dependency of the operation time
    """
    def __init__(self, general_config: GeneralSimulationConfig):
        super().__init__()
        self.__start_time = general_config.start  # s
        self.__exchange_current_decrease = 0  # p.u.
        self.__resistance_increase = 0  # p.u.
        self.__timestep = general_config.timestep

    # def calculate_resistance_increase(self, hydrogen_state: HydrogenState) -> None:
    #     relative_time_h = (hydrogen_state.time - self.__start_time) / 3600  # s -> h
    #     resistance_increase_over_total_lifetime = 0.12699  # Ohm cm²
    #     resistance_bol = 0.18416  # reference resistance at atmospheric pressure, i = 2 A/cm², T = 80°C
    #     rel_resistance_increase = resistance_increase_over_total_lifetime / resistance_bol
    #     lifetime = 80000  # h
    #     self.__resistance_increase = rel_resistance_increase / lifetime * relative_time_h

    def calculate_resistance_increase(self, state: ElectrolyzerState) -> None:
        current_dens = state.current_density
        timestep_h = self.__timestep / 3600  # s -> h
        resistance_increase_over_total_lifetime = 0.12699  # Ohm cm²
        resistance_bol = 0.18416  # reference resistance at atmospheric pressure, i = 2 A/cm², T = 80°C
        rel_resistance_increase = resistance_increase_over_total_lifetime / resistance_bol
        lifetime = 80000  # h
        if current_dens == 0:
            self.__resistance_increase += 0
        else:
            self.__resistance_increase += rel_resistance_increase / lifetime * timestep_h

    def calculate_exchange_current_dens_decrease(self, state: ElectrolyzerState):
        """
        Calculation of exchange current density decrease dependent on time the electrolyzer cell is in operation.
        based on paper: "Polymer electrolyte membrane water electrolysis: Restraining degradation in the presence
        of fluctuating power"  by Rakousky, Christoph
        year: 2017
        :param state:
        :return: none
        """
        relative_time_h = (state.time - self.__start_time) / 3600  # s -> h
        self.__exchange_current_decrease = 54.34 / 100 * m.exp(-0.007806 * relative_time_h) + 45.56 / 100

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def get_exchange_current_dens_decrease(self) -> float:
        return self.__exchange_current_decrease

    def reset(self, state: ElectrolyzerState) -> None:
        pass

    def close(self) -> None:
        pass






import numpy as np

from simses.commons.log import Logger
from simses.commons.utils.utilities import check
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class M2bAcDcConverter(AcDcConverter):
    """
    Efficiency curve of Modular Multilevel Battery Converter by mBee / STABL.
    Based on Master thesis of Markus Foerstl.

    Valid for a one-phase system with eight battery modules as described in:

    F\"orstl et al.
    "The Efficiency and Profitability of the Modular Multilevel Battery for Frequency Containment Reserve".
    In: 14th International Renewable Energy Storage Conference 2020 (IRES 2020), pp. 80-85
    doi: 10.2991/ahe.k.210202.012


    Functionality explained in
    Helling et al. "The AC lithium_ion –
    A novel approach for integrating batteries into AC systems" In: Electrical Power
    and Energy Systems 104 (2019) 150-158.
    DOI: 10.1016/j.ijepes.2018.06.047

    Inverter equation based on
    G.A. Rampinelli, A. Krenzinger, and F. Chenlo Romero.
    "Mathematical models for efficiency of inverters used in grid connected photovoltaic systems".
    In: Renewable and Sustainable Energy Reviews 34 (June 2014), pp. 578–587.
    doi: 10.1016/j.rser.2014.03.047.

    efficiency = p / (p + k0 + k1*p k2*p^2),
        where p = p_ac / p_max
    """

    # Rampinelli Fit
    __K0 = 0.003189
    __K1 = -0.007566
    __K2 = 0.009991

    __VOLUMETRIC_POWER_DENSITY = 143 * 1e6  # W / m3
    __GRAVIMETRIC_POWER_DENSITY = 17000  # W/kg
    __SPECIFIC_SURFACE_AREA = 0.0001  # in m2 / W  # TODO add exact values
    # Exemplary value from:
    # (https://www.iisb.fraunhofer.de/en/research_areas/vehicle_electronics/dcdc_converters/High_Power_Density.html)
    # ( https://www.apcuk.co.uk/app/uploads/2018/02/PE_Full_Pack.pdf )

    def __init__(self, max_power: float):
        super().__init__(max_power)
        self.__log: Logger = Logger(type(self).__name__)

    def to_ac(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power < 0.0:
            res = power / self.__get_efficiency(power)
        return res

    def to_dc(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power > 0.0:
            res = power * self.__get_efficiency(power)
        return res

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power > 0.0:
            p = (abs(dc_power) * self.max_power * (1 + self.__K1)) / (self.__K2 * abs(dc_power) - self.max_power)
            q = (abs(dc_power) * self.__K0 * self.max_power ** 2) / (self.__K2 * abs(dc_power) - self.max_power)
            res = max(0.0, -p / 2 + np.sqrt((p / 2) ** 2 - q))
        return res

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power < 0.0:
            p = (self.max_power + self.__K1 * self.max_power) / self.__K2
            q = (self.max_power ** 2 * self.__K0 - abs(dc_power) * self.max_power) / self.__K2
            res = min(0.0, -(-p / 2 + np.sqrt((p / 2) ** 2 - q)))
        return res

    def __get_efficiency(self, power: float) -> float:
        power_ratio = abs(power) / self.max_power
        if power_ratio < 0.0 or power_ratio > 1.0:
            raise Exception('Power ratio is not possible: ' + str(power_ratio))
        return power_ratio / (power_ratio + self.__K0 + self.__K1 * power_ratio + self.__K2 * power_ratio**2)

    @property
    def volume(self) -> float:
        return self.max_power / self.__VOLUMETRIC_POWER_DENSITY

    @property
    def mass(self):
        return self.max_power / self.__GRAVIMETRIC_POWER_DENSITY

    @property
    def surface_area(self) -> float:
        return self.max_power * self.__SPECIFIC_SURFACE_AREA

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        return M2bAcDcConverter(max_power)

    def close(self) -> None:
        self.__log.close()



from simses.system.dc_coupling.dc_coupler import DcCoupling
from simses.system.dc_coupling.generation.pv_dc_generation import PvDcGeneration
from simses.system.dc_coupling.load.dc_radio_ups_load import DCRadioUPSLoad


class USPDCCoupling(DcCoupling):

    def __init__(self, charging_power: float, generation_power: float):
        super().__init__(DCRadioUPSLoad(charging_power), PvDcGeneration(generation_power))


# if RadioUPSLoad is a profile this will be important

'''
class UPSDCCoupling(DcCoupling):

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig, file_name: str):
        super(UPSDCCoupling, self).__init__(RadioUPSLoad(config, profile_config, file_name), NoDcGeneration())
'''


from simses.commons.profile.power.power_profile import PowerProfile


class ConstantPowerProfile(PowerProfile):

    """
    ConstantPowerProfile delivers a constant power value over time.
    """

    def __init__(self, power: float = 0.0, scaling_factor: float = 1.0):
        """
        Constructor of ConstantPowerProfile

        Parameters
        ----------
        power :
            constant power applied in W, default: 0.0
        scaling_factor :
            linear scaling of values, default: 1.0
        """
        super().__init__()
        self.__power: float = power
        self.__scaling_factor: float = scaling_factor

    def next(self, time: float) -> float:
        return self.__power * self.__scaling_factor

    def close(self) -> None:
        pass



import math

import numpy

from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.membrane_model import \
    PemMembraneModel
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.pressure_model import \
    PemPressureModel


class PemElectricalModel:

    R_ELE = 0.096  # Ohm cm²
    P_H2_REF = 1  # bar
    P_O2_REF = 1  # bar

    def __init__(self, membrane_model: PemMembraneModel, pressure_model: PemPressureModel, parameters):
        self.__membrane_model = membrane_model
        self.__pressure_model = pressure_model
        self.__parameters = parameters

    def get_cell_voltage(self, current_dens: float, state: ElectrolyzerState):
        u_nernst = self.__nernst_voltage(state, current_dens)
        u_act = self.__activation_voltage(state, current_dens)
        r_ohm = self.get_ohm_resistance(state.temperature)
        u_ohm = r_ohm * (1 + state.resistance_increase) * current_dens  # V
        return u_nernst + u_act + u_ohm

    def get_current_density(self, power_dens_cell: float, state: ElectrolyzerState):
        stack_temperature = state.temperature  # °C
        u_nernst = self.__nernst_voltage_for_current_calc(state)
        r_ohm = self.get_ohm_resistance(state.temperature) * (1.0 + state.resistance_increase)
        # coefficients for solving quadratic equation
        A = r_ohm + self.__parameters.q1 * self.__parameters.p20
        B = u_nernst + self.__parameters.p10 + self.__parameters.q17 * self.__parameters.p11 * stack_temperature - self.__alpha(stack_temperature) \
            * numpy.log(self.__i_0(stack_temperature) * state.exchange_current_decrease)
        C = - power_dens_cell
        # calculation of currentdensity
        if power_dens_cell == 0:
            return 0
        else:
            return (- B + (B ** 2 - 4 * A * C) ** (1 / 2)) / (2 * A)

    def __reversible_voltage(self, stack_temperature, q3: float=1, q4: float=1, q5: float=1, q6: float=1) -> float:
        return 1.5184 - q3 * 1.5421 * 10 ** (-3) * (stack_temperature + 273.15) + q4 * 9.523 * 10 ** (-5) * \
               (stack_temperature + 273.15) * numpy.log(q5 * (stack_temperature + 273.15)) + q6 * 9.84 * 10 ** (-8) * \
               (stack_temperature + 273.15) ** 2  # from "Hydrogen science and engeneering: Materials, processed, systems.."

    def __nernst_voltage(self, state: ElectrolyzerState, current_dens) -> float:
        """ calculation of nernstvoltage for direct voltage calculation """
        stack_temperature = state.temperature  # °C
        p_h2 = self.__pressure_model.get_partial_pressure_h2(state, current_dens)
        p_o2 = self.__pressure_model.get_partial_pressure_o2(state, current_dens)
        u_rev = self.get_rev_voltage(stack_temperature)
        return u_rev + Hydrogen.IDEAL_GAS_CONST * (stack_temperature + 273.15) / \
               (2 * Hydrogen.FARADAY_CONST) * numpy.log((p_o2 / self.P_O2_REF) ** (1 / 2) *
                                                        p_h2 / self.P_H2_REF)

    def __nernst_voltage_for_current_calc(self, state: ElectrolyzerState):
        stack_temperature = state.temperature
        p_h2 = self.__pressure_model.get_partial_pressure_h2_for_current_calc(state, 0)  # current dependency of partial pressures is neglected in current dens calculation
        p_o2 = self.__pressure_model.get_partial_pressure_o2_for_current_calc(state, 0)  # current dependency of partial pressures is neglected in current dens calculation
        u_rev = self.get_rev_voltage_for_current_calc(stack_temperature)  # V
        return u_rev + self.__parameters.q7 * Hydrogen.IDEAL_GAS_CONST * (stack_temperature + 273.15) / \
               (2 * Hydrogen.FARADAY_CONST) * numpy.log((p_o2 / self.P_O2_REF) ** (1 / 2) * p_h2 / self.P_H2_REF)


    def __activation_voltage(self, state: ElectrolyzerState, current_dens: float) -> float:
        if current_dens == 0:
            return 0  # V
        else:
            stack_temperature = state.temperature
            exchange_current_decrease = state.exchange_current_decrease
            return self.__alpha(stack_temperature) * math.log(current_dens / (self.__i_0(stack_temperature) * exchange_current_decrease))  # V

    def __alpha(self, stack_temperature):
        """ Calculation of catalyst activity for anode and cathode combined dependent on stack temperature in °C"""
        return 0.6627 * math.exp(-0.187 * stack_temperature) + 0.02934 * math.exp(-0.00454 * stack_temperature)  # V

    def __i_0(self, stack_temperature) -> float:
        """ Calculation of exchangecurrent density dependent on stack temperature in °C """
        return 0.002159 * math.exp(-0.3179 * stack_temperature) + 1.149 * 10 ** (-7) * math.exp(-0.0205 * stack_temperature)  # A

    def get_rev_voltage_for_current_calc(self, temperature) -> float:
        """ Returns reversible Cell Voltage dependent on stack temperature in °C, but only for calculation of
         current density, use of correction parameters!!!! """
        return self.__reversible_voltage(temperature, self.__parameters.q3,
                                         self.__parameters.q4, self.__parameters.q5, self.__parameters.q6)

    def get_rev_voltage(self, temperature) -> float:
        """ Returns reversible Cell Voltage dependent on stack temperature in °C """
        return self.__reversible_voltage(temperature)

    def get_ohm_resistance(self, temperature) -> float:
        """ Returns ohmic resistance of Cell in Ohm cm² """
        return self.R_ELE + self.__membrane_model.resistance(temperature)



import pytest
from simses.commons.timeseries.average.mean_average import MeanAverage
from simses.commons.timeseries.timevalue import TimeValue


@pytest.fixture()
def uut():
    return MeanAverage()


def create_list_with(start: int, end: int) -> [TimeValue]:
    data: [TimeValue] = list()
    for value in range(start, end + 1, 1):
        data.append(TimeValue(0, value))
    return data


@pytest.mark.parametrize('start, end, result',
                         [
                             (0, 0, 0),
                             (0, 10, 5),
                             (5, 15, 10)
                         ]
                         )
def test_next(start, end, result, uut):
    data: [TimeValue] = create_list_with(start, end)
    assert abs(uut.average(data) - result) < 1e-10



from simses.commons.log import Logger
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class IntradayMarketRecharge(OperationStrategy):
    """
    If the SOC falls below a predefined lower limit or it exceeds an upper limit the \gls{bess} charges or
    discharges by trading energy on the electricity market, in particular the IDM.
    """

    __IDM_TRANSACTION_TIME = 900  # s, 15 min blocks of IDM
    __IDM_LEAD_TIME = 1800  # s, 30 min lead time for IDM transactions

    def __init__(self, general_config: GeneralSimulationConfig, fcr_config: EnergyManagementConfig):
        super().__init__(OperationPriority.LOW)
        self.__log: Logger = Logger(type(self).__name__)
        self.__timestep_start = general_config.start
        if self.__IDM_TRANSACTION_TIME % general_config.timestep != 0:
            self.__log.warn('Timestep is not a least common multiple of the IDM transaction time. '
                            'Thus, the results are distorted and are not valid. '
                            'Rethink your timestep')

        self.__max_idm_power = fcr_config.max_idm_power  # W
        self.__max_fcr_power = fcr_config.max_fcr_power  # W
        self.__soc_max_system = fcr_config.max_soc  # Upper SOC of the system in p.u
        self.__soc_min_system = fcr_config.min_soc  # Lower SOC of the system in p.u
        self.__fcr_reserve = fcr_config.fcr_reserve  # h
        self.__flag_idm_charge = False  # Flag, if IDM is necessary to charge the storage system
        self.__flag_idm_discharge = False  # Flag, if IDM is necessary to discharge the storage system
        self.__idm_power_fcr = 0  # Factor for IDM transaction (0 or 1 at fcr application)
        self.__idm_wait_time = 0  # Counter until idm market starts

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        timestep = time - self.__timestep_start
        self.__timestep_start = time
        total_capacity = system_state.capacity  # Wh
        soc_min = self.__fcr_reserve * self.__max_fcr_power / total_capacity + self.__soc_min_system
        soc_max = (total_capacity - self.__fcr_reserve * self.__max_fcr_power) / total_capacity - \
                  (1 - self.__soc_max_system)
        if time % self.__IDM_TRANSACTION_TIME == 0 and\
                not (self.__flag_idm_discharge or self.__flag_idm_charge):  # check all 15 min, if IDM is necessary
            if system_state.soc < soc_min:
                self.__flag_idm_charge = True
            elif system_state.soc > soc_max:
                self.__flag_idm_discharge = True
        elif self.__flag_idm_charge or self.__flag_idm_discharge:
            self.__idm_wait_time += timestep
        else:
            return 0

        if self.__IDM_LEAD_TIME <= self.__idm_wait_time < self.__IDM_LEAD_TIME + self.__IDM_TRANSACTION_TIME:  # 15 min IDM market
            if self.__flag_idm_charge:
                self.__idm_power_fcr = 1
            else:
                self.__idm_power_fcr = -1
        elif self.__idm_wait_time >= self.__IDM_LEAD_TIME:  # IDM finished. Reset variables
            self.__flag_idm_discharge = False
            self.__flag_idm_charge = False
            self.__idm_wait_time = 0
            self.__idm_power_fcr = 0
            return 0

        return self.__max_idm_power * self.__idm_power_fcr

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.idm_power = self.__max_idm_power * self.__idm_power_fcr

    def clear(self) -> None:
        self.__idm_power_fcr = 0.0

    def close(self) -> None:
        pass



from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.technology.redox_flow.pump_algorithm.abstract_pump_algorithm import PumpAlgorithm
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class FixFlowRateStartStop(PumpAlgorithm):
    """Pump algorithm with a fixed flow rate. The value is area specific."""

    """
    A area specific flow rate of 0.2 ml/min/cm^2 corresponds to a stoichiometric factor of 2 at a current density of
    50 mA/cm^2 at the highest and lowest state-of-charge (SOC) for a SOC range between 20 % and 80 %.
    For higher current densities and broader SOC-ranges the value must be increased.
    """
    __AREA_SPECIFIC_FLOW_RATE = 0.2  # ml/(min cm^2)

    def __init__(self, stack_module: StackModule, pump: Pump, electrolyte_system: ElectrolyteSystem):
        super().__init__(pump)
        self.__log: Logger = Logger(type(self).__name__)
        self.__stack_module: StackModule = stack_module
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__hydraulic_resistance = self.__stack_module.get_hydraulic_resistance()
        self.__flow_rate_anolyte = (self.__AREA_SPECIFIC_FLOW_RATE * self.__stack_module.get_specific_cell_area() /
                                    1000000 / 60 * self.__stack_module.get_cell_per_stack() *
                                    self.__stack_module.get_serial_scale() * self.__stack_module.get_parallel_scale())
        self.__flow_rate_catholyte = (self.__AREA_SPECIFIC_FLOW_RATE * self.__stack_module.get_specific_cell_area() /
                                      1000000 / 60 * self.__stack_module.get_cell_per_stack() *
                                      self.__stack_module.get_serial_scale() * self.__stack_module.get_parallel_scale())

    def get_pressure_drop_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        pressure_drop = (self.get_flow_rate_catholyte(redox_flow_state) * self.__hydraulic_resistance *
                         self.__electrolyte_system.get_viscosity_catholyte(redox_flow_state))
        self.__log.debug('Pressure drop for catholyte: ' + str(pressure_drop / 10 ** 5) + ' bar')
        return pressure_drop

    def get_pressure_drop_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        pressure_drop = (self.get_flow_rate_anolyte(redox_flow_state) * self.__hydraulic_resistance *
                         self.__electrolyte_system.get_viscosity_anolyte(redox_flow_state))
        self.__log.debug('Pressure drop for anolyte: ' + str(pressure_drop / 10 ** 5) + ' bar')
        return pressure_drop

    def get_flow_rate_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        if redox_flow_state.power == 0:
            flow_rate = 0
        else:
            flow_rate = self.__flow_rate_anolyte
        return flow_rate

    def get_flow_rate_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        if redox_flow_state.power == 0:
            flow_rate = 0
        else:
            flow_rate = self.__flow_rate_catholyte
        return flow_rate

    def get_flow_rate_max(self) -> float:
        return max(self.__flow_rate_anolyte, self.__flow_rate_catholyte)

    def get_flow_rate_min(self) -> float:
        return min(self.__flow_rate_anolyte, self.__flow_rate_catholyte)

    def close(self):
        super().close()
        self.__log.close()
        self.__stack_module.close()
        self.__electrolyte_system.close()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit

TEMP_IDX = 0
P_A_IDX = 1
P_C_IDX = 2
POWER_DENS_IDX = 3
CURRENT_DENS_IDX = 4

IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
FARADAY_CONST: float = 96485.3321  # As/mol

LAMBDA = 25  # degree of humidification
MEM_THIKNESS = 200 * 10 ** (-4)  # cm
R_ele = 0.096  # Ohm cm²
P_O2_REF = 1  # bar
P_H2_REF = 1  # bar

EPS = np.finfo(float).eps

current_dens_lookup_data = pd.read_csv('cell_currentdensity_lookup.csv')
#current_dens_lookup_data = pd.read_csv('cell_currentdensity_lookup_no_current_influence_on_part_pressure.csv')

temperature = current_dens_lookup_data.iloc[:, TEMP_IDX]
p_anode = current_dens_lookup_data.iloc[:, P_A_IDX]
p_cathode = current_dens_lookup_data.iloc[:, P_C_IDX]
power_dens = current_dens_lookup_data.iloc[:, POWER_DENS_IDX]
current_dens = current_dens_lookup_data.iloc[:, CURRENT_DENS_IDX]

# parameters of curve fitting of activation powerdensity
p10 = -0.00705
p20 = 0.01346
p11 = -0.0001083

# data preparation
xdata = np.array([temperature, p_anode, p_cathode, power_dens])
ydata = np.array(current_dens)

def funk(x,q1, q3, q4, q5, q6, q7, q9, q10, q11, q12, q13, q14, q15, q17):
    temp = x[0, :]
    p_a = x[1, :]
    p_c = x[2, :]
    pow_dens = x[3, :]

    def alpha(temp) -> float:
        return 0.6627 * np.exp(-0.187 * temp) + 0.02934 * np.exp(-0.00454 * temp)  # V

    def i_0(temp) -> float:
        return 0.002159 * np.exp(-0.3179 * temp) + 1.149 * 10 ** (-7) * np.exp(-0.0205 * temp)  # A

    U_rev = 1 * 1.5184 - q3 * 1.5421 * 10 ** (-3) * (temp + 273.15) + q4 * 9.523 * 10 ** (-5) * (temp + 273.15) * \
            np.log(q5 * (temp + 273.15)) + q6 * 9.84 * 10 ** (-8) * (temp + 273.15) ** 2  # from "Hydrogen science and engeneering: Materials, processed, systems.."

    conduc_nafion = (0.005139 * LAMBDA - 0.00326) * \
                    np.exp(1268 * (1 / (303) - 1 / (temp + 273.15)))  # S/cm 0.14

    R_mem = MEM_THIKNESS / conduc_nafion  # Ohm cm2

    pressure_sat_h2o = q10 * 10 ** (- q11 * 2.1794 + q12 * 0.02953 * temp - q13 * 9.1837 * 10 ** (-5) * temp ** 2 +
                                    q14 * 1.4454 * 10 ** (-7) * temp ** 3)  # bar saturationpressure of water

    p_h2 = (1 + q15 * p_c) - pressure_sat_h2o  # without currentdensity dependency
    p_o2 = (1 + q9 * p_a) - pressure_sat_h2o  # without currentdensity dependency

    A = R_mem + R_ele + q1 * p20
    B = U_rev + q7 * IDEAL_GAS_CONST * (temp + 273.15) / (2 * FARADAY_CONST) * np.log(
        1 * (p_o2 / P_O2_REF) ** (1 / 2) * p_h2 / P_H2_REF) + 1 * p10 + q17 * p11 * temp - alpha(temp) * np.log(i_0(temp))
    C = - 1 * pow_dens
    return (-B + (B ** 2 - 4 * A * C) ** (1 / 2)) / (2 * A)

lb = - np.inf
ub = np.inf

lower_bounds = (lb, lb, lb, EPS, lb, lb, lb, lb, lb, lb, lb, lb, lb, lb)
#lower_bounds = lb
upper_bounds = ub

popt, pcov = curve_fit(funk, xdata, ydata, absolute_sigma=False, bounds=(lower_bounds, upper_bounds))

standart_deviation = np.sqrt(np.diag(pcov))
print(popt)
print(pcov)
print(standart_deviation)

current_ana_calc = funk(xdata, popt[0], popt[1], popt[2], popt[3], popt[4], popt[5], popt[6], popt[7], popt[8], popt[9],
                        popt[10], popt[11], popt[12], popt[13])

abs_err = current_dens - current_ana_calc

mean_abs_err = np.mean(abs(abs_err))

max_abs_err = max(abs(abs_err))

RMS = np.sqrt(sum((abs_err) ** 2) / len(abs_err))

relative_error = abs(abs_err / current_dens)

chisq = sum((abs_err) ** 2)

x1 = list(range(1, len(popt)+1))
y1 = popt
x2 = list(range(len(temperature)))
y21 = abs_err
y22 = relative_error

plt.subplot(311)
plt.ylabel('correction parameters')
plt.grid(True)
#plt.title('standart deviation of the parameters')
plt.plot(x1, y1, 'bo')
plt.xticks(x1, ('q1', 'q3', 'q4', 'q5', 'q6', 'q7', 'q9', 'q10', 'q11', 'q12', 'q13', 'q14', 'q15', 'q17'))

plt.subplot(312)
#plt.title('absolut error funktion vs look up table')
plt.ylabel('absolut error')
plt.grid(True)
plt.plot(x2, y21, 'rx')
plt.plot(x2, current_dens, 'bo')
plt.axis([-100, len(x2), min(abs_err), 0.009])

plt.subplot(313)
#plt.title('relative error funktion vs look up table')
plt.ylabel('relative error in %')
plt.grid(True)
plt.plot(x2, y22*100, 'gx')

#plt.subplot(414)
#plt.errorbar(x2, current_dens - funk(xdata, popt[0], popt[1], popt[2], popt[3], popt[4], popt[5], popt[6], popt[7], popt[8], popt[9], popt[10], popt[11], popt[12], popt[13], popt[14], popt[15], popt[16]
#                                , popt[17], popt[18], popt[19], popt[20], popt[21], popt[22], popt[23], popt[24], popt[25], popt[26], popt[27]), 0, fmt='b.')

plt.figtext(0.15, 0.8, "mean abs. error: %.6f"%mean_abs_err, fontweight="bold")
plt.figtext(0.15, 0.75, "RMS error: %.6f"%RMS, fontweight="bold")
plt.figtext(0.15, 0.7, "maximal absolut error: %.6f"%max_abs_err, fontweight="bold")
plt.grid(True)
plt.savefig('multidim current density fit - analysis')
plt.show()
#plt.savefig('multidim current density fit no curreninfluence on part pressure - standart deviation')


dict = {'parameters': popt}
df = pd.DataFrame(dict)
df.to_csv('parameters_multi_dim_current_dens_fit.csv', index=False)

# dict2 = {'temperatur':temperature, 'p anode':p_anode, 'p cathode':p_cathode, 'powerdensity':power_dens, 'currentdensity':current_dens, 'calc currentdensity':current_ana_calc, 'absolut error':abs_err}
# df2 = pd.DataFrame(dict2)
# df2.to_csv('restul best fit analytic equation currentedensity.csv', index=False)


from abc import abstractmethod

from simses.technology.hydrogen.hydrogen_storage.hydrogen_storage import HydrogenStorage


class Pipeline(HydrogenStorage):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def calculate_inijected_hydrogen(self, time_diff, hydrogen_outflow) -> None:
        pass

    @abstractmethod
    def get_injected_hydrogen(self) -> float:
        pass

    @abstractmethod
    def get_tank_pressure(self) -> float:
        pass



import math

import pandas as pd
import scipy.interpolate

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.round_18650 import RoundCell18650
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class MolicelNMC(CellType):
    """An NMC (NMC_Molicel) is a special cell type and inherited by CellType"""
    """Source: 
    Schuster, S. F., Bach, T., Fleder, E., Müller, J., Brand, M., Sextl, G., & Jossen, A. (2015). 
    Nonlinear aging characteristics of lithium-ion cells under different operational conditions. 
    Journal of Energy Storage, 1, 44–53. doi:10.1016/j.est.2015.05.003 """

    __SOC_IDX = 0

    __CELL_VOLTAGE = 3.7  # V
    __CELL_CAPACITY = 1.9  # Ah
    __MAX_VOLTAGE: float = 4.25  # V
    __MIN_VOLTAGE: float = 3.0  # V
    __MIN_TEMPERATURE: float = 273.15  # K
    __MAX_TEMPERATURE: float = 318.15  # K
    __MAX_CHARGE_RATE: float = 1.05  # 1/h
    __MAX_DISCHARGE_RATE: float = 2.1  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
    __MASS: float = 0.045  # kg per cell
    __SPECIFIC_HEAT: float = 965  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K

    __COULOMB_EFFICIENCY: float = 1.0  # p.u.

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE, __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = RoundCell18650()

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)

        self.__log: Logger = Logger(type(self).__name__)
        rint_file: str = battery_data_config.nmc_molicel_rint_file
        internal_resistance = pd.read_csv(rint_file)  # Ohm
        soc_arr = internal_resistance.iloc[:, self.__SOC_IDX]
        rint_mat_ch = internal_resistance.iloc[:, 2]
        rint_mat_dch = internal_resistance.iloc[:, 5]
        self.__rint_ch_interp1d = scipy.interpolate.interp1d(soc_arr, rint_mat_ch, kind='linear')
        self.__rint_dch_interp1d = scipy.interpolate.interp1d(soc_arr, rint_mat_dch, kind='linear')

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        '''Parameters build with ocv fitting'''
        a1 = -1.6206
        a2 = -6.9895
        a3 = 1.4458
        a4 = 1.9530
        b1 = 3.4206
        b2 = 0.8759
        k0 = 2.0127
        k1 = 2.7684
        k2 = 1.0698
        k3 = 4.1431
        k4 = -3.8417
        k5 = -0.1856
        soc = battery_state.soc

        ocv = k0 + \
              k1 / (1 + math.exp(a1 * (soc - b1))) + \
              k2 / (1 + math.exp(a2 * (soc - b2))) + \
              k3 / (1 + math.exp(a3 * (soc - 1))) +\
              k4 / (1 + math.exp(a4 * soc)) +\
              k5 * soc

        return ocv * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        soc = self.check_soc_range(battery_state.soc)
        if battery_state.is_charge:
            rint = self.__rint_ch_interp1d(soc)
        else:
            rint = self.__rint_dch_interp1d(soc)
        return float(rint) / self.get_parallel_scale() * self.get_serial_scale()

    def close(self):
        self.__log.close()
